import os
import unittest

import numpy as np

from smt.applications.mfk import MFK, NestedLHS
from smt.applications.mfkpls import MFKPLS
from smt.applications.mfkplsk import MFKPLSK
from smt.applications.moe import MOE
from smt.sampling_methods import FullFactorial


class TestSaveLoad(unittest.TestCase):
    def function_test_1d(self, x):
        x = np.reshape(x, (-1,))
        y = np.zeros(x.shape)
        y[x < 0.4] = x[x < 0.4] ** 2
        y[(x >= 0.4) & (x < 0.8)] = 3 * x[(x >= 0.4) & (x < 0.8)] + 1
        y[x >= 0.8] = np.sin(10 * x[x >= 0.8])
        return y.reshape((-1, 1))

    def test_save_load_moe(self):
        filename = "moe_save_test"
        nt = 35
        x = np.linspace(0, 1, 100)

        sampling = FullFactorial(xlimits=np.array([[0, 1]]), clip=True)
        np.random.seed(0)
        xt = sampling(nt)
        yt = self.function_test_1d(xt)

        moe1 = MOE(n_clusters=1)
        moe1.set_training_values(xt, yt)
        moe1.train()
        y_moe1 = moe1.predict_values(x)

        moe1.save(filename)

        moe2 = MOE.load(filename)
        y_moe2 = moe2.predict_values(x)

        np.testing.assert_allclose(y_moe1, y_moe2)

        os.remove(filename)

    def lf_function(self, x):
        return 0.5 * ((x * 6 - 2) ** 2) * np.sin((x * 6 - 2) * 2) + (x - 0.5) * 10.0 - 5

    def hf_function(self, x):
        return ((x * 6 - 2) ** 2) * np.sin((x * 6 - 2) * 2)

    def _setup_MFKs(self):
        xlimits = np.array([[0.0, 1.0]])
        xdoes = NestedLHS(nlevel=2, xlimits=xlimits, random_state=0)
        xt_c, xt_e = xdoes(7)
        yt_e = self.hf_function(xt_e)
        yt_c = self.lf_function(xt_c)
        x = np.linspace(0, 1, 101, endpoint=True).reshape(-1, 1)
        return (xt_c, xt_e, yt_c, yt_e, x)

    def test_save_load_mfk_mfkpls_mfkplsk(self):
        filename = "MFKs_save_test"
        MFKs = [MFK, MFKPLS, MFKPLSK]
        xt_c, xt_e, yt_c, yt_e, x = self._setup_MFKs()
        ncomp = 1
        for mfk in MFKs:
            if mfk == MFK:
                application = MFK(theta0=xt_e.shape[1] * [1.0], corr="squar_exp")
            elif mfk == MFKPLS:
                application = MFKPLS(n_comp=ncomp, theta0=ncomp * [1.0])
            else:
                application = MFKPLSK(n_comp=ncomp, theta0=ncomp * [1.0])

            application.set_training_values(xt_c, yt_c, name=0)
            application.set_training_values(xt_e, yt_e)

            application.train()
            application.save(filename)

            x = np.linspace(0, 1, 101, endpoint=True).reshape(-1, 1)
            y1 = application.predict_values(x)

            mfk2 = MFK.load(filename)
            y2 = mfk2.predict_values(x)

            np.testing.assert_allclose(y1, y2)

            os.remove(filename)


if __name__ == "__main__":
    unittest.main()



"""
Authors: Paul Saves
"""

import warnings

import numpy as np
from scipy.sparse import issparse
from sklearn.base import BaseEstimator, RegressorMixin
from sklearn.exceptions import DataConversionWarning, NotFittedError
from sklearn.utils.validation import check_array, check_X_y, check_is_fitted
from smt.surrogate_models import QP
from smt.surrogate_models.surrogate_model import SurrogateModel

from sklearn.utils import Tags


class ScikitLearnAdapter(RegressorMixin, BaseEstimator):
    """
    Adapter to wrap SMT SurrogateModel instances with scikit-learn compatible API.

    All keyword arguments passed to __init__ (besides model_cls) are treated
    as hyperparameters for the surrogate model.
    """

    def get_tags(self):
        return Tags(
            requires_y=True,
            input_tags={"sparse": False},  # indicates sparse support
            X_types=["2darray"],
        )

    def __init__(self, model_cls=QP, **model_kwargs):
        # model_cls must be a subclass of SurrogateModel
        try:
            if not issubclass(model_cls, SurrogateModel):
                raise ValueError(
                    f"model_cls must be a subclass of SurrogateModel, got {model_cls}"
                )
            # store parameters as attributes for get_params/set_params
        except TypeError:
            model_cls = QP

        self.model_cls = model_cls
        for name, value in model_kwargs.items():
            setattr(self.model_cls, name, value)
            setattr(self, name, value)

    def fit(self, X, y):
        X, y = check_X_y(X, y, accept_sparse=False)
        if issparse(X):
            X = X.toarray()
        # build surrogate model with current hyperparameters
        params_mod = self.get_params(deep=True)
        model_cls = params_mod.pop("model_cls")
        params = dict(model_cls.__dict__)
        # assemble kwargs for the surrogate
        surrogate_kwargs = {k: params[k] for k in params}
        surrogate2_kwargs = self.__dict__
        common_dict = {
            k: surrogate_kwargs[k]
            for k in surrogate_kwargs.keys() & surrogate2_kwargs.keys()
        }

        self.model_ = model_cls(**common_dict)
        X_arr = np.atleast_2d(np.asarray(X))
        self.n_features_in_ = X_arr.shape[1]
        y_arr = np.asarray(y)
        if y_arr.ndim == 1:
            y_arr = y_arr[:, None]
        elif y_arr.ndim > 1:
            warnings.warn(
                "DataConversionWarning('A column-vector y"
                " was passed when a 1d array was expected",
                category=DataConversionWarning,
            )
        if len(X_arr.flatten()) < 1:
            raise ValueError(
                f"0 feature(s) (shape={X.shape}) while a minimum of 1 is required."
            )

        self.model_.set_training_values(X_arr, y_arr)
        self.model_.train()
        return self

    def predict(self, X):
        check_is_fitted(self, "model_")
        X = check_array(X, accept_sparse=False)
        if issparse(X):
            X = X.toarray()
        X_arr = np.asarray(X)
        if X_arr.ndim == 1:
            raise ValueError(
                f"Expected 2D array, got 1D array instead with shape {X_arr.shape}. "
                "Reshape your data using X.reshape(1, -1)"
            )
        X_arr = np.atleast_2d(np.asarray(X))
        self.n_features_in_ = X_arr.shape[1]

        if len(X_arr.flatten()) < 1:
            raise ValueError(
                f"0 feature(s) (shape={X.shape}) while a minimum of 1 is required."
            )
        try:
            y_pred = self.model_.predict_values(X_arr)
        except AttributeError:
            raise NotFittedError("not fitted")
        # flatten if single-output
        if y_pred.shape[1] == 1:
            return y_pred.ravel()
        return y_pred



"""
Author: Mohamed Amine Bouhlel <mbouhlel@umich.edu>

This package is distributed under New BSD license.
"""

import unittest

try:
    import matplotlib

    matplotlib.use("Agg")
    NO_MATPLOTLIB = False
except ImportError:
    NO_MATPLOTLIB = True

import numpy as np

from smt.applications import VFM
from smt.examples.rans_crm_wing.rans_crm_wing import (
    get_rans_crm_wing,
)
from smt.problems import WaterFlow, WaterFlowLFidelity
from smt.sampling_methods import LHS
from smt.utils.misc import compute_relative_error
from smt.utils.silence import Silence
from smt.utils.sm_test_case import SMTestCase


def setupCRM(LF_candidate="QP", Bridge_candidate="KRG", type_bridge="Additive"):
    xt, yt, _ = get_rans_crm_wing()

    optionsLF = {}
    optionsB = {}

    M = VFM(
        type_bridge=type_bridge,
        name_model_LF=LF_candidate,
        name_model_bridge=Bridge_candidate,
        X_LF=xt,
        y_LF=yt,
        X_HF=xt,
        y_HF=yt,
        options_LF=optionsLF,
        options_bridge=optionsB,
    )

    return M, xt


class TestVFM(SMTestCase):
    def test_vfm(self):
        # Problem set up
        ndim = 8
        ntest = 500
        ndoeLF = int(10 * ndim)
        ndoeHF = int(3)
        funLF = WaterFlowLFidelity(ndim=ndim)
        funHF = WaterFlow(ndim=ndim)
        deriv1 = True
        deriv2 = True
        LF_candidate = "QP"
        Bridge_candidate = "KRG"
        type_bridge = "Multiplicative"
        optionsLF = {}
        optionsB = {
            "theta0": [1e-2] * ndim,
            "print_prediction": False,
            "deriv": False,
            "hyper_opt": "Cobyla",
        }

        # Construct low/high fidelity data and validation points
        sampling = LHS(xlimits=funLF.xlimits, criterion="m", random_state=42)
        xLF = sampling(ndoeLF)
        yLF = funLF(xLF)
        if deriv1:
            dy_LF = np.zeros((ndoeLF, 1))
            for i in range(ndim):
                yd = funLF(xLF, kx=i)
                dy_LF = np.concatenate((dy_LF, yd), axis=1)

        sampling = LHS(xlimits=funHF.xlimits, criterion="m", random_state=43)
        xHF = sampling(ndoeHF)
        yHF = funHF(xHF)
        if deriv2:
            dy_HF = np.zeros((ndoeHF, 1))
            for i in range(ndim):
                yd = funHF(xHF, kx=i)
                dy_HF = np.concatenate((dy_HF, yd), axis=1)

        xtest = sampling(ntest)
        ytest = funHF(xtest)
        dytest = np.zeros((ntest, ndim))
        for i in range(ndim):
            dytest[:, i] = funHF(xtest, kx=i).T

        # Initialize VFM
        vfm = VFM(
            type_bridge=type_bridge,
            name_model_LF=LF_candidate,
            name_model_bridge=Bridge_candidate,
            X_LF=xLF,
            y_LF=yLF,
            X_HF=xHF,
            y_HF=yHF,
            options_LF=optionsLF,
            options_bridge=optionsB,
            dy_LF=dy_LF,
            dy_HF=dy_HF,
        )

        # Prediction of the validation points
        rms_error = compute_relative_error(vfm, xtest, ytest)
        self.assert_error(rms_error, 0.0, 3e-1)

    @staticmethod
    def run_vfm_example(self):
        import matplotlib.pyplot as plt
        import numpy as np

        from smt.applications import VFM
        from smt.problems import WaterFlow, WaterFlowLFidelity
        from smt.sampling_methods import LHS

        # Problem set up
        ndim = 8
        ntest = 500
        ndoeLF = int(10 * ndim)
        ndoeHF = int(3)
        funLF = WaterFlowLFidelity(ndim=ndim)
        funHF = WaterFlow(ndim=ndim)
        deriv1 = True
        deriv2 = True
        LF_candidate = "QP"
        Bridge_candidate = "KRG"
        type_bridge = "Multiplicative"
        optionsLF = {}
        optionsB = {
            "theta0": [1e-2] * ndim,
            "print_prediction": False,
            "deriv": False,
            "hyper_opt": "Cobyla",
        }

        # Construct low/high fidelity data and validation points
        sampling = LHS(xlimits=funLF.xlimits, criterion="m")
        xLF = sampling(ndoeLF)
        yLF = funLF(xLF)
        if deriv1:
            dy_LF = np.zeros((ndoeLF, 1))
            for i in range(ndim):
                yd = funLF(xLF, kx=i)
                dy_LF = np.concatenate((dy_LF, yd), axis=1)

        sampling = LHS(xlimits=funHF.xlimits, criterion="m")
        xHF = sampling(ndoeHF)
        yHF = funHF(xHF)
        if deriv2:
            dy_HF = np.zeros((ndoeHF, 1))
            for i in range(ndim):
                yd = funHF(xHF, kx=i)
                dy_HF = np.concatenate((dy_HF, yd), axis=1)

        xtest = sampling(ntest)
        ytest = funHF(xtest)
        dytest = np.zeros((ntest, ndim))
        for i in range(ndim):
            dytest[:, i] = funHF(xtest, kx=i).T

        # Initialize the extension VFM
        M = VFM(
            type_bridge=type_bridge,
            name_model_LF=LF_candidate,
            name_model_bridge=Bridge_candidate,
            X_LF=xLF,
            y_LF=yLF,
            X_HF=xHF,
            y_HF=yHF,
            options_LF=optionsLF,
            options_bridge=optionsB,
            dy_LF=dy_LF,
            dy_HF=dy_HF,
        )

        # Prediction of the validation points
        y = M.predict_values(x=xtest)

        plt.figure()
        plt.plot(ytest, ytest, "-.")
        plt.plot(ytest, y, ".")
        plt.xlabel(r"$y$ True")
        plt.ylabel(r"$y$ prediction")
        plt.show()

    def test_KRG_KRG_additive(self):
        with Silence():
            M, xt = setupCRM(
                LF_candidate="KRG", Bridge_candidate="KRG", type_bridge="Additive"
            )

        with Silence():
            yp = M.predict_values(np.atleast_2d(xt[0]))
            dyp = M.predict_derivatives(np.atleast_2d(xt[0]), kx=0)
        self.assert_error(yp, np.array([[0.015368, 0.367424]]), atol=2e-2, rtol=3e-2)
        self.assert_error(
            dyp,
            np.array([[0.31539861, 7.30043182]]),
            atol=5e-1,
            rtol=5e-2,
        )

    def test_QP_KRG_additive(self):
        with Silence():
            M, xt = setupCRM(
                LF_candidate="QP", Bridge_candidate="KRG", type_bridge="Additive"
            )

        with Silence():
            yp = M.predict_values(np.atleast_2d(xt[0]))
            dyp = M.predict_derivatives(np.atleast_2d(xt[0]), kx=0)

        self.assert_error(
            yp, np.array([[0.01535549, 0.36720433]]), atol=1e-2, rtol=1e-2
        )
        self.assert_error(
            dyp, np.array([[0.27030465, 5.89010571]]), atol=5e-1, rtol=5e-2
        )

    def test_KRG_KRG_mult(self):
        with Silence():
            M, xt = setupCRM(
                LF_candidate="KRG", Bridge_candidate="KRG", type_bridge="Multiplicative"
            )

        with Silence():
            yp = M.predict_values(np.atleast_2d(xt[0]))
            dyp = M.predict_derivatives(np.atleast_2d(xt[0]), kx=0)

        self.assert_error(
            yp, np.array([[0.01536264, 0.36713384]]), atol=2e-2, rtol=3e-2
        )
        self.assert_error(
            dyp, np.array([[0.29039224, 7.16030401]]), atol=5e-1, rtol=5e-2
        )

    def test_QP_KRG_mult(self):
        with Silence():
            M, xt = setupCRM(
                LF_candidate="QP", Bridge_candidate="KRG", type_bridge="Multiplicative"
            )

        with Silence():
            yp = M.predict_values(np.atleast_2d(xt[0]))
            dyp = M.predict_derivatives(np.atleast_2d(xt[0]), kx=0)

        self.assert_error(
            yp, np.array([[0.01536423, 0.36742269]]), atol=3e-1, rtol=1e-2
        )
        self.assert_error(
            dyp, np.array([[0.19562493, 5.39545518]]), atol=3e-1, rtol=1e-2
        )


if __name__ == "__main__":
    unittest.main()



# -*- coding: utf-8 -*-
"""
Created on Fri May 04 10:26:49 2018

@author: Mostafa Meliani <melimostafa@gmail.com>
Multi-Fidelity co-Kriging: recursive formulation with autoregressive model of
order 1 (AR1)
Adapted on January 2021 by Andres Lopez-Lopera to the new SMT version
"""

import warnings
from copy import deepcopy

import numpy as np
from scipy import linalg
from scipy.linalg import solve_triangular
from scipy.spatial.distance import cdist
from sklearn.cross_decomposition import PLSRegression as pls

from smt.design_space import ensure_design_space
from smt.sampling_methods import LHS
from smt.surrogate_models.krg_based import (
    KrgBased,
    MixIntKernelType,
    compute_n_param,
)
from smt.utils.kriging import (
    componentwise_distance,
    compute_X_cont,
    cross_distances,
    cross_levels,
    differences,
    gower_componentwise_distances,
)
from smt.utils.misc import standardization


class NestedLHS(object):
    def __init__(self, nlevel, xlimits=None, design_space=None, random_state=None):
        """
        Constructor where values of options can be passed in.

        Parameters
        ----------
        nlevel : integer.
            The number of design of experiments to be built

        xlimits : ndarray
            The interval of the domain in each dimension with shape (nx, 2)

        design_space : DesignSpace
            The design space with type and bounds for every design variable

        random_state : Numpy RandomState object or seed number which controls random draws

        """
        self.nlevel = nlevel
        self.random_state = random_state
        if xlimits is None and design_space is None:
            raise ValueError(
                "Either xlimits or design_space should be specified to have bounds for the sampling."
            )
        elif xlimits is not None and design_space is not None:
            raise ValueError(
                "Use either design_space for mixed inputs or xlimits for continuous one. \
                    Please avoid overspecification."
            )
        elif xlimits is not None:
            self.design_space = ensure_design_space(xlimits=xlimits)
        else:
            self.design_space = design_space

    def __call__(self, nb_samples_hifi):
        """
        Builds nlevel nested design of experiments of dimension dim and size n_samples.
        Each doe is built with an optmized LHS procedure.
        The method builds the highest level first and
        nested properties are ensured by deleting the nearest neighbours in lower levels of fidelity.

        Parameters
        ----------

        nb_samples_hifi: The number of samples of the highest fidelity model.
            nb_samples_fi(n-1) = 2 * nb_samples_fi(n)


        Returns
        ------

        list of length nlevel of design of experiemnts from low to high fidelity level.
        """
        nt = []
        for i in range(self.nlevel, 0, -1):
            nt.append(pow(2, i - 1) * nb_samples_hifi)

        if len(nt) != self.nlevel:
            raise ValueError("nt must be a list of nlevel elements")
        if not np.allclose(np.sort(nt)[::-1], nt):
            raise ValueError("nt must be a list of decreasing integers")

        doe = []
        p0 = LHS(
            xlimits=np.array(self.design_space.get_unfolded_num_bounds()),
            criterion="ese",
            random_state=self.random_state,
        )
        p0nt0 = p0(nt[0])
        if self.design_space:
            p0nt0, _ = self.design_space.correct_get_acting(p0nt0)
            p0nt0, _ = self.design_space.fold_x(p0nt0)

        doe.append(p0nt0)

        for i in range(1, self.nlevel):
            p = LHS(
                xlimits=np.array(self.design_space.get_unfolded_num_bounds()),
                criterion="ese",
                random_state=self.random_state,
            )
            pnti = p(nt[i])
            if self.design_space:
                pnti, _ = self.design_space.correct_get_acting(pnti)
                pnti, _ = self.design_space.fold_x(pnti)

            doe.append(pnti)

        for i in range(1, self.nlevel)[::-1]:
            ind = []
            d = cdist(doe[i], doe[i - 1], "euclidean")
            for j in range(doe[i].shape[0]):
                dj = np.sort(d[j, :])
                k = dj[0]
                ll = (np.where(d[j, :] == k))[0][0]
                m = 0
                while ll in ind:
                    m = m + 1
                    k = dj[m]
                    ll = (np.where(d[j, :] == k))[0][0]
                ind.append(ll)

            doe[i - 1] = np.delete(doe[i - 1], ind, axis=0)
            doe[i - 1] = np.vstack((doe[i - 1], doe[i]))
        return doe


class MFK(KrgBased):
    def _initialize(self):
        super()._initialize()
        declare = self.options.declare
        declare(
            "rho_regr",
            "constant",
            values=("constant", "linear", "quadratic"),
            desc="Regression function type for rho",
        )
        declare(
            "optim_var",
            False,
            types=bool,
            values=(True, False),
            desc="If True, the variance at HF samples is forced to zero",
        )
        declare(
            "propagate_uncertainty",
            True,
            types=bool,
            values=(True, False),
            desc="If True, the variance cotribution of lower fidelity levels are considered",
        )
        self.name = "MFK"

        self.X2_norma = {}
        self.X2_offset = {}
        self.X2_scale = {}

    def _differences(self, X, Y):
        """
        Compute the distances
        """
        return differences(X, Y)

    def _check_list_structure(self, X, y):
        """
        checks if the data structure is compatible with MFK.
        sets class attributes such as (number of levels of Fidelity, training points in each level, ...)

        Arguments :
        X : list of arrays, each array corresponds to a fidelity level. starts from lowest to highest
        y : same as X
        """

        if not isinstance(X, list):
            nlevel = 1
            X = [X]
        else:
            nlevel = len(X)

        if not isinstance(y, list):
            y = [y]

        if len(X) != len(y):
            raise ValueError("X and y must have the same length.")

        n_samples = np.zeros(nlevel, dtype=int)
        n_features = np.zeros(nlevel, dtype=int)
        n_samples_y = np.zeros(nlevel, dtype=int)
        for i in range(nlevel):
            n_samples[i], n_features[i] = X[i].shape
            if i > 1 and n_features[i] != n_features[i - 1]:
                raise ValueError("All X must have the same number of columns.")
            y[i] = np.asarray(y[i]).ravel()[:, np.newaxis]
            n_samples_y[i] = y[i].shape[0]
            if n_samples[i] != n_samples_y[i]:
                raise ValueError("X and y must have the same number of rows.")

        self.nx = n_features[0]
        self.nt_all = n_samples
        self.nlvl = nlevel
        self.ny = y[0].shape[1]
        self.X = X[:]
        self.y = y[:]

    def _new_train(self):
        """
        Overrides KrgBased implementation
        Trains the Multi-Fidelity model
        """
        self._corr_params = None

        self._new_train_init()
        theta0 = self.options["theta0"].copy()
        noise0 = self.options["noise0"].copy()

        for lvl in range(self.nlvl):
            self._new_train_iteration(lvl)
            self.options["theta0"] = theta0
            self.options["noise0"] = noise0

        self._reinterpolate(lvl)

    def _new_train_init(self):
        if self.name in ["MFKPLS", "MFKPLSK"]:
            _pls = pls(self.options["n_comp"])

            # As of sklearn 0.24.1 PLS with zeroed outputs raises an exception
            # while sklearn 0.23 returns zeroed x_rotations
            # For now the try/except below is a workaround to restore the 0.23 behaviour
            try:
                # PLS is done on the highest fidelity identified by the key None
                self.m_pls = _pls.fit(
                    self.training_points[None][0][0].copy(),
                    self.training_points[None][0][1].copy(),
                )
                self.coeff_pls = self.m_pls.x_rotations_
            except StopIteration:
                self.coeff_pls = np.zeros(
                    self.training_points[None][0][0].shape[1], self.options["n_comp"]
                )

        xt = []
        yt = []
        i = 0
        while self.training_points.get(i, None) is not None:
            xt.append(self.training_points[i][0][0])
            yt.append(self.training_points[i][0][1])
            i = i + 1
        xt.append(self.training_points[None][0][0])
        yt.append(self.training_points[None][0][1])

        self._check_list_structure(xt, yt)
        self._check_param()
        X = self.X
        y = self.y

        if self.is_continuous:
            (
                _,
                _,
                self.X_offset,
                self.y_mean,
                self.X_scale,
                self.y_std,
            ) = standardization(np.concatenate(xt, axis=0), np.concatenate(yt, axis=0))
            self.cat_features = [False] * np.shape(np.concatenate(xt, axis=0))[1]
        else:
            (
                _,
                _,
                self.X_offset,
                self.y_mean,
                self.X_scale,
                self.y_std,
            ) = standardization(np.concatenate(xt, axis=0), np.concatenate(yt, axis=0))
            _, self.cat_features = compute_X_cont(
                np.concatenate(xt, axis=0), self.design_space
            )
            self.X_offset[self.cat_features] *= 0
            self.X_scale[self.cat_features] *= 0
            self.X_scale[self.cat_features] += 1

        nlevel = self.nlvl

        # initialize lists
        self.optimal_noise_all = nlevel * [0]
        self.D_all = nlevel * [0]
        self.F_all = nlevel * [0]
        self.p_all = nlevel * [0]
        self.q_all = nlevel * [0]
        self.optimal_rlf_value = nlevel * [0]
        self.optimal_par = nlevel * [{}]
        self.optimal_theta = nlevel * [0]
        self.X_norma_all = [(x - self.X_offset) / self.X_scale for x in X]
        self.y_norma_all = [(f - self.y_mean) / self.y_std for f in y]

        self.X_norma_all = [(x - self.X_offset) / self.X_scale for x in X]

        self.y_norma_all = [(f - self.y_mean) / self.y_std for f in y]

        if not (self.is_continuous):
            x2t, _, _ = self.design_space.unfold_x(np.concatenate(xt, axis=0))
            _, _, self.X2_offset, _, self.X2_scale, _ = standardization(
                x2t, np.concatenate(yt, axis=0)
            )

    def _new_train_iteration(self, lvl):
        n_samples = self.nt_all
        self.options["noise0"] = np.array([self.options["noise0"][lvl]]).flatten()
        self.options["theta0"] = self.options["theta0"][lvl, :]

        self.X_norma = self.X_norma_all[lvl]
        self.y_norma = self.y_norma_all[lvl]

        if self.options["eval_noise"]:
            if self.options["use_het_noise"]:
                # hetGP works with unique design variables
                (
                    self.X_norma,
                    self.index_unique,  # do we need to store it?
                    self.nt_reps,  # do we need to store it?
                ) = np.unique(
                    self.X_norma, return_inverse=True, return_counts=True, axis=0
                )
                self.nt_all[lvl] = self.X_norma.shape[0]

                # computing the mean of the output per unique design variable (see Binois et al., 2018)
                y_norma_unique = []
                for i in range(self.nt_all[lvl]):
                    y_norma_unique.append(np.mean(self.y_norma[self.index_unique == i]))
                y_norma_unique = np.array(y_norma_unique).reshape(-1, 1)

                # pointwise sensible estimates of the noise variances (see Ankenman et al., 2010)
                self.optimal_noise = self.options["noise0"] * np.ones(self.nt_all[lvl])
                for i in range(self.nt_all[lvl]):
                    diff = self.y_norma[self.index_unique == i] - y_norma_unique[i]
                    if np.sum(diff**2) != 0.0:
                        self.optimal_noise[i] = np.std(diff, ddof=1) ** 2
                self.optimal_noise = self.optimal_noise / self.nt_reps
                self.optimal_noise_all[lvl] = self.optimal_noise
                self.y_norma = y_norma_unique

                self.X_norma_all[lvl] = self.X_norma
                self.y_norma_all[lvl] = self.y_norma
        else:
            self.optimal_noise = self.options["noise0"] / self.y_std**2
            self.optimal_noise_all[lvl] = self.optimal_noise

        # Calculate matrix of distances D between samples
        if self.is_continuous:
            self.D_all[lvl] = cross_distances(self.X_norma)
        else:
            _, is_acting = self.design_space.correct_get_acting(self.X[lvl])
            D_lvl, ij_lvl, X, _ = gower_componentwise_distances(
                X=self.X[lvl],
                x_is_acting=is_acting,
                design_space=self.design_space,
                hierarchical_kernel=self.options["hierarchical_kernel"],
            )
            self.Lij, self.n_levels = cross_levels(
                X=self.X[lvl], ij=ij_lvl, design_space=self.design_space
            )
            self.D_all[lvl] = D_lvl, ij_lvl

        # Regression matrix and parameters
        self.F_all[lvl] = self._regression_types[self.options["poly"]](self.X_norma)
        self.p_all[lvl] = self.F_all[lvl].shape[1]

        # Concatenate the autoregressive part for levels > 0
        if lvl > 0:
            F_rho = self._regression_types[self.options["rho_regr"]](self.X_norma)
            self.q_all[lvl] = F_rho.shape[1]

            if self.is_continuous:
                self.F_all[lvl] = np.hstack(
                    (
                        F_rho
                        * np.dot(
                            self._predict_intermediate_values(
                                self.X_norma, lvl, descale=False
                            ),
                            np.ones((1, self.q_all[lvl])),
                        ),
                        self.F_all[lvl],
                    )
                )
            else:
                self.F_all[lvl] = np.hstack(
                    (
                        F_rho
                        * np.dot(
                            self._predict_intermediate_values(
                                self.X[lvl], lvl, descale=False
                            ),
                            np.ones((1, self.q_all[lvl])),
                        ),
                        self.F_all[lvl],
                    )
                )

        else:
            self.q_all[lvl] = 0

        n_samples_F_i = self.F_all[lvl].shape[0]

        if n_samples_F_i != n_samples[lvl]:
            raise Exception(
                "Number of rows in F and X do not match. Most "
                "likely something is going wrong with the "
                "regression model."
            )

        if int(self.p_all[lvl] + self.q_all[lvl]) >= n_samples_F_i:
            raise Exception(
                (
                    "Ordinary least squares problem is undetermined "
                    "n_samples=%d must be greater than the regression"
                    " model size p+q=%d."
                )
                % (n_samples_F_i, self.p_all[lvl] + self.q_all[lvl])
            )

        # Determine Gaussian Process model parameters
        self.F = self.F_all[lvl]
        D, self.ij = self.D_all[lvl]
        self._lvl = lvl
        self.nt = self.nt_all[lvl]
        self.q = self.q_all[lvl]
        self.p = self.p_all[lvl]
        self.kplsk_second_loop = False
        (
            self.optimal_rlf_value[lvl],
            self.optimal_par[lvl],
            self.optimal_theta[lvl],
        ) = self._optimize_hyperparam(D)
        if self.options["eval_noise"] and not self.options["use_het_noise"]:
            tmp_list = self.optimal_theta[lvl]
            self.optimal_theta[lvl] = tmp_list[:-1]
            self.optimal_noise = tmp_list[-1]
            self.optimal_noise_all[lvl] = self.optimal_noise
        del self.y_norma, self.D, self.optimal_noise

    def _reinterpolate(self, lvl):
        if self.options["eval_noise"] and self.options["optim_var"]:
            X = self.X
            for lvl in range(self.nlvl - 1):
                self.set_training_values(
                    X[lvl], self._predict_intermediate_values(X[lvl], lvl + 1), name=lvl
                )
            self.set_training_values(
                X[-1], self._predict_intermediate_values(X[-1], self.nlvl)
            )
            self.options["eval_noise"] = False
            self._new_train()
            self.options["eval_noise"] = True

    def _componentwise_distance(self, dx, opt=0):
        d = componentwise_distance(
            dx, self.options["corr"], self.nx, power=self.options["pow_exp_power"]
        )
        return d

    def _predict_intermediate_values(self, X, lvl, descale=True):
        """
        Evaluates the model at a set of points.
        Used for training the model at level lvl.
        Allows to relax the order problem.

        Arguments
        ---------
        x : np.ndarray [n_evals, dim]
            Evaluation point input variable values
        lvl : level at which the prediction is made

        Returns
        -------
        y : np.ndarray
            Evaluation point output variable values
        """
        n_eval, _ = X.shape
        #        if n_features_X != self.n_features:
        #            raise ValueError("Design must be an array of n_features columns.")

        # Calculate kriging mean and variance at level 0
        mu = np.zeros((n_eval, lvl))

        if not (self.is_continuous):
            X_usc = X
        if descale and self.is_continuous:
            X = (X - self.X_offset) / self.X_scale

        f = self._regression_types[self.options["poly"]](X)
        f0 = self._regression_types[self.options["poly"]](X)

        beta = self.optimal_par[0]["beta"]
        if self.is_continuous:
            dx = self._differences(X, Y=self.X_norma_all[0])
            d = self._componentwise_distance(dx)
            self.corr.theta = self.optimal_theta[0]
            r_ = self.corr(d).reshape(n_eval, self.nt_all[0])
        else:
            _, x_is_acting = self.design_space.correct_get_acting(X_usc)
            _, y_is_acting = self.design_space.correct_get_acting(self.X[0])
            dx, _ = gower_componentwise_distances(
                X=X_usc,
                x_is_acting=x_is_acting,
                design_space=self.design_space,
                hierarchical_kernel=self.options["hierarchical_kernel"],
                y=np.copy(self.X[0]),
                y_is_acting=y_is_acting,
            )
            d = componentwise_distance(
                dx,
                self.options["corr"],
                self.nx,
                power=self.options["pow_exp_power"],
                theta=None,
                return_derivative=False,
            )
            _, ij = cross_distances(X_usc, self.X[0])
            Lij, _ = cross_levels(
                X=X_usc, ij=ij, design_space=self.design_space, y=self.X[0]
            )
            self.ij = ij
            if self.options["categorical_kernel"] == MixIntKernelType.CONT_RELAX:
                Xpred, _, _ = self.design_space.unfold_x(X_usc)
                Xpred_norma = (Xpred - self.X2_offset) / self.X2_scale
                # Get pairwise componentwise L1-distances to the input training set
                dx = differences(Xpred_norma, Y=self.X2_norma[str(0)].copy())
            r_ = self._matrix_data_corr(
                corr=self.options["corr"],
                design_space=self.design_space,
                power=self.options["pow_exp_power"],
                theta=self.optimal_theta[0],
                theta_bounds=self.options["theta_bounds"],
                dx=dx,
                Lij=Lij,
                n_levels=self.n_levels,
                cat_features=self.cat_features,
                cat_kernel=self.options["categorical_kernel"],
                x=X_usc,
            ).reshape(n_eval, self.nt_all[0])

        gamma = self.optimal_par[0]["gamma"]

        # Scaled predictor
        mu[:, 0] = (np.dot(f, beta) + np.dot(r_, gamma)).ravel()

        # Calculate recursively kriging mean and variance at level i
        for i in range(1, lvl):
            g = self._regression_types[self.options["rho_regr"]](X)

            if self.is_continuous:
                dx = self._differences(X, Y=self.X_norma_all[i])
                d = self._componentwise_distance(dx)
                self.corr.theta = self.optimal_theta[i]
                r_ = self.corr(d).reshape(n_eval, self.nt_all[i])
            else:
                _, x_is_acting = self.design_space.correct_get_acting(X_usc)
                _, y_is_acting = self.design_space.correct_get_acting(self.X[i])
                dx, _ = gower_componentwise_distances(
                    X_usc,
                    x_is_acting=x_is_acting,
                    design_space=self.design_space,
                    hierarchical_kernel=self.options["hierarchical_kernel"],
                    y=np.copy(self.X[i]),
                    y_is_acting=y_is_acting,
                )
                d = componentwise_distance(
                    dx,
                    self.options["corr"],
                    self.nx,
                    power=self.options["pow_exp_power"],
                    theta=None,
                    return_derivative=False,
                )
                _, ij = cross_distances(X_usc, self.X[i])
                Lij, _ = cross_levels(
                    X=X_usc, ij=ij, design_space=self.design_space, y=self.X[i]
                )
                self.ij = ij
                if self.options["categorical_kernel"] == MixIntKernelType.CONT_RELAX:
                    Xpred, _, _ = self.design_space.unfold_x(X_usc)
                    Xpred_norma = (Xpred - self.X2_offset) / self.X2_scale
                    # Get pairwise componentwise L1-distances to the input training set
                    dx = differences(Xpred_norma, Y=self.X2_norma[str(i)].copy())
                r_ = self._matrix_data_corr(
                    corr=self.options["corr"],
                    design_space=self.design_space,
                    power=self.options["pow_exp_power"],
                    theta=self.optimal_theta[i],
                    theta_bounds=self.options["theta_bounds"],
                    dx=dx,
                    Lij=Lij,
                    n_levels=self.n_levels,
                    cat_features=self.cat_features,
                    cat_kernel=self.options["categorical_kernel"],
                    x=X_usc,
                ).reshape(n_eval, self.nt_all[i])

            f = np.vstack((g.T * mu[:, i - 1], f0.T))
            beta = self.optimal_par[i]["beta"]
            gamma = self.optimal_par[i]["gamma"]
            # scaled predictor
            mu[:, i] = (np.dot(f.T, beta) + np.dot(r_, gamma)).ravel()

        # scaled predictor
        if descale:
            mu = mu * self.y_std + self.y_mean

        return mu[:, -1].reshape((n_eval, 1))

    def _predict_values(self, X, is_acting=None):
        """
        Evaluates the model at a set of points.

        Arguments
        ---------
        x : np.ndarray [n_evals, dim]
            Evaluation point input variable values

        Returns
        -------
        y : np.ndarray
            Evaluation point output variable values
        """

        return self._predict_intermediate_values(X, self.nlvl)

    def _predict_variances(
        self, X: np.ndarray, is_acting=None, is_ri=False
    ) -> np.ndarray:
        """
        Evaluates the model at a set of points.

        Arguments
        ---------
        x : np.ndarray [n_evals, dim]
            Evaluation point input variable values

        Returns
        -------
        y : np.ndarray
            Evaluation point output variable values
        """
        return self.predict_variances_all_levels(X)[0][:, -1]

    def predict_variances_all_levels(self, X, is_acting=None):
        # def predict_variances_all_levels(self, X):
        """
        Evaluates the model at a set of points.

        Arguments
        ---------
        x : np.ndarray [n_evals, dim]
            Evaluation point input variable values

        Returns
        -------
        y : np.ndarray
            Evaluation point output variable values
        """
        # Initialization X = atleast_2d(X)
        nlevel = self.nlvl
        sigma2_rhos = []
        n_eval, n_features_X = X.shape
        #        if n_features_X != self.n_features:
        #            raise ValueError("Design must be an array of n_features columns.")

        X = (X - self.X_offset) / self.X_scale

        # Calculate kriging mean and variance at level 0
        mu = np.zeros((n_eval, nlevel))

        f = self._regression_types[self.options["poly"]](X)
        f0 = self._regression_types[self.options["poly"]](X)

        # Get regression function and correlation
        F = self.F_all[0]
        C = self.optimal_par[0]["C"]

        beta = self.optimal_par[0]["beta"]
        Ft = solve_triangular(C, F, lower=True)

        if self.is_continuous:
            dx = self._differences(X, Y=self.X_norma_all[0])
            d = self._componentwise_distance(dx)
            self.corr.theta = self.optimal_theta[0]
            r_ = self.corr(d).reshape(n_eval, self.nt_all[0])
        else:
            _, y_is_acting = self.design_space.correct_get_acting(self.X[0])
            _, x_is_acting = self.design_space.correct_get_acting(X)
            dx, _ = gower_componentwise_distances(
                X,
                x_is_acting=x_is_acting,
                design_space=self.design_space,
                hierarchical_kernel=self.options["hierarchical_kernel"],
                y=np.copy(self.X[0]),
                y_is_acting=y_is_acting,
            )
            d = componentwise_distance(
                dx,
                self.options["corr"],
                self.nx,
                power=self.options["pow_exp_power"],
                theta=None,
                return_derivative=False,
            )
            _, ij = cross_distances(X, self.X[0])
            Lij, _ = cross_levels(
                X=X, ij=ij, design_space=self.design_space, y=self.X[0]
            )
            self.ij = ij
            if self.options["categorical_kernel"] == MixIntKernelType.CONT_RELAX:
                Xpred, _, _ = self.design_space.unfold_x(X)
                Xpred_norma = (Xpred - self.X2_offset) / self.X2_scale
                # Get pairwise componentwise L1-distances to the input training set
                dx = differences(Xpred_norma, Y=self.X2_norma[str(0)].copy())
            r_ = self._matrix_data_corr(
                corr=self.options["corr"],
                design_space=self.design_space,
                power=self.options["pow_exp_power"],
                theta=self.optimal_theta[0],
                theta_bounds=self.options["theta_bounds"],
                dx=dx,
                Lij=Lij,
                n_levels=self.n_levels,
                cat_features=self.cat_features,
                cat_kernel=self.options["categorical_kernel"],
                x=X,
            ).reshape(n_eval, self.nt_all[0])

        gamma = self.optimal_par[0]["gamma"]

        # Scaled predictor
        mu[:, 0] = (np.dot(f, beta) + np.dot(r_, gamma)).ravel()

        self.sigma2_rho = nlevel * [None]
        MSE = np.zeros((n_eval, nlevel))
        r_t = solve_triangular(C, r_.T, lower=True)
        G = self.optimal_par[0]["G"]

        u_ = solve_triangular(G.T, f.T - np.dot(Ft.T, r_t), lower=True)
        sigma2 = self.optimal_par[0]["sigma2"] / self.y_std**2
        MSE[:, 0] = sigma2 * (
            # 1 + self.optimal_noise_all[0] - (r_t ** 2).sum(axis=0) + (u_ ** 2).sum(axis=0)
            1 - (r_t**2).sum(axis=0) + (u_**2).sum(axis=0)
        )

        # Calculate recursively kriging variance at level i
        for i in range(1, nlevel):
            F = self.F_all[i]
            C = self.optimal_par[i]["C"]

            g = self._regression_types[self.options["rho_regr"]](X)

            if self.is_continuous:
                dx = self._differences(X, Y=self.X_norma_all[i])
                d = self._componentwise_distance(dx)
                self.corr.theta = self.optimal_theta[i]
                r_ = self.corr(d).reshape(n_eval, self.nt_all[i])
            else:
                _, y_is_acting = self.design_space.correct_get_acting(self.X[i])
                _, x_is_acting = self.design_space.correct_get_acting(X)
                dx, _ = gower_componentwise_distances(
                    X,
                    x_is_acting=x_is_acting,
                    design_space=self.design_space,
                    hierarchical_kernel=self.options["hierarchical_kernel"],
                    y=np.copy(self.X[i]),
                    y_is_acting=y_is_acting,
                )
                d = componentwise_distance(
                    dx,
                    self.options["corr"],
                    self.nx,
                    power=self.options["pow_exp_power"],
                    theta=None,
                    return_derivative=False,
                )
                _, ij = cross_distances(X, self.X[i])
                Lij, _ = cross_levels(
                    X=X, ij=ij, design_space=self.design_space, y=self.X[i]
                )
                self.ij = ij
                if self.options["categorical_kernel"] == MixIntKernelType.CONT_RELAX:
                    Xpred, _, _ = self.design_space.unfold_x(X)
                    Xpred_norma = (Xpred - self.X2_offset) / self.X2_scale
                    # Get pairwise componentwise L1-distances to the input training set
                    dx = differences(Xpred_norma, Y=self.X2_norma[str(i)].copy())
                r_ = self._matrix_data_corr(
                    corr=self.options["corr"],
                    design_space=self.design_space,
                    power=self.options["pow_exp_power"],
                    theta=self.optimal_theta[i],
                    theta_bounds=self.options["theta_bounds"],
                    dx=dx,
                    Lij=Lij,
                    n_levels=self.n_levels,
                    cat_features=self.cat_features,
                    cat_kernel=self.options["categorical_kernel"],
                    x=X,
                ).reshape(n_eval, self.nt_all[i])

            f = np.vstack((g.T * mu[:, i - 1], f0.T))

            Ft = solve_triangular(C, F, lower=True)
            yt = solve_triangular(C, self.y_norma_all[i], lower=True)
            r_t = solve_triangular(C, r_.T, lower=True)
            G = self.optimal_par[i]["G"]
            beta = self.optimal_par[i]["beta"]

            # scaled predictor
            sigma2 = self.optimal_par[i]["sigma2"] / self.y_std**2
            q = self.q_all[i]
            u_ = solve_triangular(G.T, f - np.dot(Ft.T, r_t), lower=True)
            sigma2_rho = np.dot(
                g,
                sigma2 * linalg.inv(np.dot(G.T, G))[:q, :q]
                + np.dot(beta[:q], beta[:q].T),
            )
            sigma2_rho = (sigma2_rho * g).sum(axis=1)
            sigma2_rhos.append(sigma2_rho)

            if self.name in ["MFKPLS", "MFKPLSK"]:
                p = self.p_all[i]
                Q_ = (np.dot((yt - np.dot(Ft, beta)).T, yt - np.dot(Ft, beta)))[0, 0]
                MSE[:, i] = (
                    # sigma2_rho * MSE[:, i - 1]
                    +Q_
                    / (2 * (self.nt_all[i] - p - q))
                    # * (1 + self.optimal_noise_all[i] - (r_t ** 2).sum(axis=0))
                    * (1 - (r_t**2).sum(axis=0))
                    + sigma2 * (u_**2).sum(axis=0)
                )
            else:
                MSE[:, i] = sigma2 * (
                    # 1 + self.optimal_noise_all[i] - (r_t ** 2).sum(axis=0) + (u_ ** 2).sum(axis=0)
                    1 - (r_t**2).sum(axis=0) + (u_**2).sum(axis=0)
                )  # + sigma2_rho * MSE[:, i - 1]
            if self.options["propagate_uncertainty"]:
                MSE[:, i] = MSE[:, i] + sigma2_rho * MSE[:, i - 1]

        # scaled predictor
        MSE *= self.y_std**2

        return MSE, sigma2_rhos

    def _predict_derivatives(self, x, kx):
        """
        Evaluates the derivatives at a set of points.

        Arguments
        ---------
        x : np.ndarray [n_evals, dim]
            Evaluation point input variable values
        kx : int
            The 0-based index of the input variable with respect to which derivatives are desired.

        Returns
        -------
        y : np.ndarray*self.y_std/self.X_scale[kx])
            Derivative values.
        """

        lvl = self.nlvl
        # Initialization

        n_eval, n_features_x = x.shape
        x = (x - self.X_offset) / self.X_scale

        dy_dx = np.zeros((n_eval, lvl))

        if self.options["corr"] != "squar_exp":
            raise ValueError(
                "The derivative is only available for square exponential kernel"
            )
        if self.options["poly"] == "constant":
            df = np.zeros([n_eval, 1])
        elif self.options["poly"] == "linear":
            df = np.zeros((n_eval, self.nx + 1))
            df[:, 1:] = 1
        else:
            raise ValueError(
                "The derivative is only available for ordinary kriging or "
                + "universal kriging using a linear trend"
            )

        df0 = deepcopy(df)

        if self.options["rho_regr"] != "constant":
            raise ValueError(
                "The derivative is only available for regression rho constant"
            )
        # Get pairwise componentwise L1-distances to the input training set
        dx = self._differences(x, Y=self.X_norma_all[0])
        d = self._componentwise_distance(dx)
        # Compute the correlation function
        self.corr.theta = self.optimal_theta[0]
        r_ = self.corr(d).reshape(n_eval, self.nt_all[0])

        # Beta and gamma = R^-1(y-FBeta)
        beta = self.optimal_par[0]["beta"]
        gamma = self.optimal_par[0]["gamma"]

        df_dx = np.dot(df, beta)
        d_dx = x[:, kx].reshape((n_eval, 1)) - self.X_norma_all[0][:, kx].reshape(
            (1, self.nt_all[0])
        )

        theta = self._get_theta(0)
        dy_dx[:, 0] = np.ravel((df_dx - 2 * theta[kx] * np.dot(d_dx * r_, gamma)))

        # Calculate recursively derivative at level i
        for i in range(1, lvl):
            g = self._regression_types[self.options["rho_regr"]](x)
            dx = self._differences(x, Y=self.X_norma_all[i])
            d = self._componentwise_distance(dx)
            self.corr.theta = self.optimal_theta[i]
            r_ = self.corr(d).reshape(n_eval, self.nt_all[i])
            df = np.vstack((g.T * dy_dx[:, i - 1], df0.T))

            beta = self.optimal_par[i]["beta"]
            gamma = self.optimal_par[i]["gamma"]

            df_dx = np.dot(df.T, beta)
            d_dx = x[:, kx].reshape((n_eval, 1)) - self.X_norma_all[i][:, kx].reshape(
                (1, self.nt_all[i])
            )

            theta = self._get_theta(i)

            # scaled predictor
            dy_dx[:, i] = np.ravel(df_dx - 2 * theta[kx] * np.dot(d_dx * r_, gamma))

        deriv = dy_dx[:, -1] * self.y_std / self.X_scale[kx]
        return deriv.reshape((x.shape[0], self.ny))

    def _get_theta(self, i):
        return self.optimal_theta[i]

    def _check_param(self):
        """
        Overrides KrgBased implementation
        This function checks some parameters of the model.
        """

        if self.name in ["MFKPLS", "MFKPLSK"]:
            d = self.options["n_comp"]
        else:
            d = self.nx

        if self.options["corr"] == "act_exp":
            raise ValueError("act_exp correlation function must be used with MGP")

        if self.name in ["MFKPLS"]:
            if self.options["corr"] not in ["squar_exp", "abs_exp"]:
                raise ValueError(
                    "MFKPLS only works with a squared exponential or an absolute exponential kernel"
                )
        elif self.name in ["MFKPLSK"]:
            if self.options["corr"] not in ["squar_exp"]:
                raise ValueError(
                    "MFKPLSK only works with a squared exponential kernel (until we prove the contrary)"
                )
        # noise0 may be a list of noise values for various fi levels with various length
        max_noise = np.max([np.max(row) for row in self.options["noise0"]])
        if (self.options["eval_noise"] or max_noise > 1e-12) and self.options[
            "hyper_opt"
        ] == "TNC":
            self.options["hyper_opt"] = "Cobyla"
            warnings.warn(
                "TNC not available yet for noise handling. Switching to Cobyla"
            )
        n_param = d

        if self.options["categorical_kernel"] is not None:
            if self.options["categorical_kernel"] == MixIntKernelType.CONT_RELAX:
                n_comp = self.options["n_comp"] if "n_comp" in self.options else None
                mat_dim = (
                    self.options["cat_kernel_comps"]
                    if "cat_kernel_comps" in self.options
                    else None
                )

                n_param = compute_n_param(
                    self.design_space,
                    self.options["categorical_kernel"],
                    d,
                    n_comp,
                    mat_dim,
                )
            else:
                raise ValueError(
                    "Only the continuous relaxation kernel is available with multi-fidelity"
                )

        if isinstance(self.options["theta0"], np.ndarray):
            if self.options["theta0"].shape != (self.nlvl, n_param):
                raise ValueError(
                    "the dimensions of theta0 %s should coincide to the number of dim %s"
                    % (self.options["theta0"].shape, (self.nlvl, n_param))
                )
        else:
            if len(self.options["theta0"]) != n_param:
                if len(self.options["theta0"]) == 1:
                    self.options["theta0"] *= np.ones((self.nlvl, n_param))
                elif len(self.options["theta0"]) == self.nlvl:
                    self.options["theta0"] = np.array(self.options["theta0"]).reshape(
                        -1, 1
                    )
                    self.options["theta0"] *= np.ones((1, n_param))
                else:
                    raise ValueError(
                        "the length of theta0 (%s) should be equal to the number of dim (%s) \
                            or levels of fidelity (%s)."
                        % (len(self.options["theta0"]), n_param, self.nlvl)
                    )
            else:
                self.options["theta0"] *= np.ones((self.nlvl, 1))

        if len(self.options["noise0"]) != self.nlvl:
            if len(self.options["noise0"]) == 1:
                self.options["noise0"] = self.nlvl * [self.options["noise0"]]
            else:
                raise ValueError(
                    "the length of noise0 (%s) should be equal to the number of levels of fidelity (%s)."
                    % (len(self.options["noise0"]), self.nlvl)
                )

        for i in range(self.nlvl):
            if self.options["use_het_noise"]:
                if len(self.X[i]) == len(np.unique(self.X[i])):
                    if len(self.options["noise0"][i]) != self.nt_all[i]:
                        if len(self.options["noise0"][i]) == 1:
                            self.options["noise0"][i] *= np.ones(self.nt_all[i])
                        else:
                            raise ValueError(
                                "for the level of fidelity %s, the length of noise0 (%s) should be equal to \
                                    the number of observations (%s)."
                                % (i, len(self.options["noise0"][i]), self.nt_all[i])
                            )
            else:
                if np.size(self.options["noise0"][i]) != 1:
                    raise ValueError(
                        "for the level of fidelity %s, the length of noise0 (%s) should be equal to one."
                        % (i, len(self.options["noise0"][i]))
                    )



"""
Created on Tue Oct 12 10:48:01 2021
@author: psaves
"""

import itertools
import os
import unittest

import numpy as np

try:
    import matplotlib

    matplotlib.use("Agg")
    NO_MATPLOTLIB = False
except ImportError:
    NO_MATPLOTLIB = True

from smt.applications.mixed_integer import (
    MixedIntegerContext,
    MixedIntegerKrigingModel,
    MixedIntegerSamplingMethod,
)
from smt.design_space import (
    CategoricalVariable,
    DesignSpace,
    FloatVariable,
    IntegerVariable,
    OrdinalVariable,
)
from smt.problems import HierarchicalGoldstein, HierarchicalNeuralNetwork, Sphere
from smt.sampling_methods import LHS
from smt.surrogate_models import (
    KPLS,
    KPLSK,
    KRG,
    QP,
    MixHrcKernelType,
    MixIntKernelType,
)


class TestMixedInteger(unittest.TestCase):
    def test_periodic_mixed(self):
        # Objective function
        def f_obj(X):
            """
            s01 objective

            Parameters
            ----------
            point: array_like
                point to evaluate
            """
            design_space = DesignSpace(
                [
                    FloatVariable(0.0, 1.0),
                    CategoricalVariable(
                        values=[
                            "1",
                            "2",
                            "3",
                            "4",
                            "5",
                            "6",
                            "7",
                            "8",
                            "9",
                            "10",
                            "11",
                            "12",
                            "13",
                        ]
                    ),
                ]
            )

            PI = 3.14159265358979323846
            fail = False
            x = X[0]
            #  caté 1
            c = str(design_space.decode_values(X, i_dv=1)[1])
            x = np.abs(x)
            c1 = c == "1"
            c2 = c == "2"
            c3 = c == "3"
            c4 = c == "4"
            c5 = c == "5"
            c6 = c == "6"
            c7 = c == "7"
            c8 = c == "8"
            c9 = c == "9"
            c10 = c == "10"
            c11 = c == "11"
            c12 = c == "12"
            c13 = c == "13"

            if np.size(c1) == (
                np.sum(c1)
                + np.sum(c2)
                + np.sum(c3)
                + np.sum(c4)
                + np.sum(c5)
                + np.sum(c6)
                + np.sum(c7)
                + np.sum(c8)
                + np.sum(c9)
                + np.sum(c10)
                + np.sum(c11)
                + np.sum(c12)
                + np.sum(c13)
            ):
                u = (
                    1 * c1
                    + 2 * c2
                    + 3 * c3
                    + 4 * c4
                    + 5 * c5
                    + 6 * c6
                    + 7 * c7
                    + 8 * c8
                    + 9 * c9
                    + 10 * c10
                    + 11 * c11
                    + 12 * c12
                    + 13 * c13
                )
                hu = (PI * (0.4 + u / 15)) * (c10 + c11 + c12 + c13) - u / 20
                y = np.cos(3.5 * PI * x + hu)
            else:
                print("type error")
                print(X)
                fail = True
            return (y, fail)

        n_doe = 98

        design_space = DesignSpace(
            [
                FloatVariable(0.0, 1.0),
                CategoricalVariable(
                    values=[
                        "1",
                        "2",
                        "3",
                        "4",
                        "5",
                        "6",
                        "7",
                        "8",
                        "9",
                        "10",
                        "11",
                        "12",
                        "13",
                    ]
                ),
            ]
        )

        sampling = MixedIntegerSamplingMethod(
            LHS,
            design_space,
            criterion="ese",
            random_state=42,
            output_in_folded_space=True,
        )
        xdoe = sampling(n_doe)

        y_doe = [f_obj(xdoe[i])[0] for i in range(len(xdoe))]
        # Surrogate
        for m in MixIntKernelType:
            sm = MixedIntegerKrigingModel(
                surrogate=KRG(
                    design_space=design_space,
                    categorical_kernel=m,
                    theta0=[0.01],
                    hyper_opt="Cobyla",
                    corr="squar_sin_exp",
                    n_start=5,
                ),
            )

            sm.set_training_values(xdoe, np.array(y_doe))
            if m in [MixIntKernelType.EXP_HOMO_HSPHERE, MixIntKernelType.CONT_RELAX]:
                with self.assertRaises(ValueError):
                    sm.train()
            else:
                sm.train()

    def test_krg_mixed_3D(self):
        design_space = DesignSpace(
            [
                FloatVariable(-10, 10),
                CategoricalVariable(["blue", "red", "green"]),
                IntegerVariable(-10, 10),
            ]
        )

        mixint = MixedIntegerContext(design_space)

        sm = mixint.build_kriging_model(KRG(hyper_opt="Cobyla", print_prediction=False))
        sampling = mixint.build_sampling_method()

        fun = Sphere(ndim=3)
        xt = sampling(20)
        yt = fun(xt)
        sm.set_training_values(xt, yt)
        sm.train()

        eq_check = True
        for i in range(xt.shape[0]):
            if abs(float(xt[i, :][2]) - int(float(xt[i, :][2]))) > 10e-8:
                eq_check = False
            if not (xt[i, :][1] == 0 or xt[i, :][1] == 1 or xt[i, :][1] == 2):
                eq_check = False
        self.assertTrue(eq_check)

    def test_krg_mixed_3D_bad_regr(self):
        design_space = DesignSpace(
            [
                FloatVariable(-10, 10),
                CategoricalVariable(["blue", "red", "green"]),
                IntegerVariable(-10, 10),
            ]
        )

        mixint = MixedIntegerContext(design_space)
        with self.assertRaises(ValueError):
            _sm = mixint.build_kriging_model(
                KRG(hyper_opt="Cobyla", print_prediction=False, poly="linear")
            )

    def test_qp_mixed_2D_INT(self):
        design_space = DesignSpace(
            [
                FloatVariable(-10, 10),
                IntegerVariable(-10, 10),
            ]
        )

        mixint = MixedIntegerContext(design_space)
        sm = mixint.build_surrogate_model(QP(print_prediction=False))
        sampling = mixint.build_sampling_method()

        fun = Sphere(ndim=2)
        xt = sampling(10)
        yt = fun(xt)
        sm.set_training_values(xt, yt)
        sm.train()

        eq_check = True
        for i in range(xt.shape[0]):
            if abs(float(xt[i, :][1]) - int(float(xt[i, :][1]))) > 10e-8:
                eq_check = False
        self.assertTrue(eq_check)

    def test_compute_unfolded_dimension(self):
        design_space = DesignSpace(
            [
                FloatVariable(0, 1),
                CategoricalVariable(["A", "B"]),
            ]
        )
        assert design_space.get_unfolded_num_bounds().shape[0] == 3

    def test_unfold_with_enum_mask(self):
        x = np.array([[1.5, 1], [1.5, 0], [1.5, 1]])
        expected = [[1.5, 0, 1], [1.5, 1, 0], [1.5, 0, 1]]

        design_space = DesignSpace(
            [
                FloatVariable(1, 2),
                CategoricalVariable(["A", "B"]),
            ]
        )
        x_unfolded, _, _ = design_space.unfold_x(x)
        self.assertListEqual(expected, x_unfolded.tolist())

    def test_unfold_with_enum_mask_with_enum_first(self):
        x = np.array([[1, 1.5], [0, 1.5], [1, 1.5]])
        expected = [[0, 1, 1.5], [1, 0, 1.5], [0, 1, 1.5]]

        design_space = DesignSpace(
            [
                CategoricalVariable(["A", "B"]),
                FloatVariable(1, 2),
            ]
        )
        x_unfolded, _, _ = design_space.unfold_x(x)
        self.assertListEqual(expected, x_unfolded.tolist())

    def test_fold_with_enum_index(self):
        x = np.array([[1.5, 0, 1], [1.5, 1, 0], [1.5, 0, 1]])
        expected = [[1.5, 1], [1.5, 0], [1.5, 1]]

        design_space = DesignSpace(
            [
                FloatVariable(1, 2),
                CategoricalVariable(["A", "B"]),
            ]
        )
        x_folded, _ = design_space.fold_x(x)
        self.assertListEqual(expected, x_folded.tolist())

    def test_fold_with_enum_index_with_list(self):
        expected = [[1.5, 1]]
        x = np.array([1.5, 0, 1])

        design_space = DesignSpace(
            [
                FloatVariable(1, 2),
                CategoricalVariable(["A", "B"]),
            ]
        )
        x_folded, _ = design_space.fold_x(x)
        self.assertListEqual(expected, x_folded.tolist())
        x = [1.5, 0, 1]
        x_folded, _ = design_space.fold_x(x)
        self.assertListEqual(expected, x_folded.tolist())

    def test_cast_to_enum_value(self):
        design_space = DesignSpace(
            [
                FloatVariable(0, 4),
                CategoricalVariable(["blue", "red"]),
            ]
        )
        x = np.zeros((5, 2))
        x[:, 1] = [1, 1, 0, 1, 0]
        decoded = design_space.decode_values(x, i_dv=1)
        expected = ["red", "red", "blue", "red", "blue"]
        self.assertListEqual(expected, decoded)

    def test_unfolded_xlimits_type(self):
        design_space = DesignSpace(
            [
                FloatVariable(-5, 5),
                CategoricalVariable(["2", "3"]),
                CategoricalVariable(["4", "5"]),
                IntegerVariable(0, 2),
            ]
        )
        sampling = MixedIntegerSamplingMethod(LHS, design_space, criterion="ese")
        doe = sampling(10)
        self.assertEqual((10, 4), doe.shape)

    def test_unfold_xlimits_with_continuous_limits(self):
        design_space = DesignSpace(
            [
                FloatVariable(-5, 5),
                CategoricalVariable(["blue", "red"]),
                CategoricalVariable(["short", "medium", "long"]),
                IntegerVariable(0, 2),
            ]
        )

        unfolded_limits = design_space.get_unfolded_num_bounds()
        self.assertEqual(
            np.array_equal(
                [[-5, 5], [0, 1], [0, 1], [0, 1], [0, 1], [0, 1], [0, 2]],
                unfolded_limits,
            ),
            True,
        )

    def test_unfold_xlimits_with_continuous_limits_and_ordinal_values(self):
        design_space = DesignSpace(
            [
                FloatVariable(-5, 5),
                CategoricalVariable(["blue", "red"]),
                CategoricalVariable(["short", "medium", "long"]),
                OrdinalVariable(["0", "3", "4"]),
            ]
        )

        unfolded_limits = design_space.get_unfolded_num_bounds()
        self.assertEqual(
            np.array_equal(
                [[-5, 5], [0, 1], [0, 1], [0, 1], [0, 1], [0, 1], [0, 2]],
                unfolded_limits,
            ),
            True,
        )

    def test_cast_to_discrete_values(self):
        design_space = DesignSpace(
            [
                FloatVariable(-5, 5),
                CategoricalVariable(["blue", "red"]),
                CategoricalVariable(["short", "medium", "long"]),
                IntegerVariable(0, 4),
            ]
        )

        x = np.array([[2.6, 0.3, 0.5, 0.25, 0.45, 0.85, 3.1]])
        np.testing.assert_allclose(
            np.array([[2.6, 0, 1, 0, 0, 1, 3]]),
            design_space.correct_get_acting(x)[0],
            atol=1e-9,
        )

    def test_cast_to_discrete_values_with_smooth_rounding_ordinal_values(self):
        x = np.array([[2.6, 0.3, 0.5, 0.25, 0.45, 0.85, 1.1]])
        design_space = DesignSpace(
            [
                FloatVariable(-5, 5),
                CategoricalVariable(["blue", "red"]),
                CategoricalVariable(["short", "medium", "long"]),
                OrdinalVariable(["0", "2", "4"]),
            ]
        )

        np.testing.assert_allclose(
            np.array([[2.6, 0, 1, 0, 0, 1, 1]]),
            design_space.correct_get_acting(x)[0],
            atol=1e-9,
        )

    def test_cast_to_discrete_values_with_hard_rounding_ordinal_values(self):
        x = np.array([[2.6, 0.3, 0.5, 0.25, 0.45, 0.85, 0.9]])
        design_space = DesignSpace(
            [
                FloatVariable(-5, 5),
                CategoricalVariable(["blue", "red"]),
                CategoricalVariable(["short", "medium", "long"]),
                OrdinalVariable(["0", "4"]),
            ]
        )

        np.testing.assert_allclose(
            np.array([[2.6, 0, 1, 0, 0, 1, 1]]),
            design_space.correct_get_acting(x)[0],
            atol=1e-9,
        )

    def test_cast_to_discrete_values_with_non_integer_ordinal_values(self):
        x = np.array([[2.6, 0.3, 0.5, 0.25, 0.45, 0.85, 0.8]])
        design_space = DesignSpace(
            [
                FloatVariable(-5, 5),
                CategoricalVariable(["blue", "red"]),
                CategoricalVariable(["short", "medium", "long"]),
                OrdinalVariable(["0", "3.5"]),
            ]
        )
        np.testing.assert_allclose(
            np.array([[2.6, 0, 1, 0, 0, 1, 1]]),
            design_space.correct_get_acting(x)[0],
            atol=1e-9,
        )

    @unittest.skipIf(NO_MATPLOTLIB, "Matplotlib not installed")
    def test_examples(self):
        self.run_mixed_integer_lhs_example()
        self.run_mixed_integer_qp_example()
        self.run_mixed_integer_context_example()
        self.run_hierarchical_variables_Goldstein()
        self.run_mixed_discrete_design_space_example()
        self.run_mixed_gower_example()
        self.run_mixed_homo_gaussian_example()
        self.run_mixed_homo_hyp_example()
        # FIXME: this test should belong to smt_design_space_ext
        # but at the moment run_* code is used here to generate doc here in smt
        # if HAS_DESIGN_SPACE_EXT:
        #     self.run_mixed_cs_example()
        #     self.run_hierarchical_design_space_example()  # works only with config space impl

    def run_mixed_integer_lhs_example(self):
        import matplotlib.pyplot as plt
        import numpy as np
        from matplotlib import colors

        from smt.applications.mixed_integer import MixedIntegerSamplingMethod
        from smt.design_space import (
            CategoricalVariable,
            DesignSpace,
            FloatVariable,
        )
        from smt.sampling_methods import LHS

        float_var = FloatVariable(0, 4)
        cat_var = CategoricalVariable(["blue", "red"])

        design_space = DesignSpace(
            [
                float_var,
                cat_var,
            ]
        )

        num = 40
        design_space.seed = 42
        samp = MixedIntegerSamplingMethod(
            LHS, design_space, criterion="ese", random_state=design_space.seed
        )
        x, x_is_acting = samp(num, return_is_acting=True)

        cmap = colors.ListedColormap(cat_var.values)
        plt.scatter(x[:, 0], np.zeros(num), c=x[:, 1], cmap=cmap)
        plt.show()

    def run_mixed_integer_qp_example(self):
        import matplotlib.pyplot as plt
        import numpy as np

        from smt.applications.mixed_integer import MixedIntegerSurrogateModel
        from smt.design_space import DesignSpace, IntegerVariable
        from smt.surrogate_models import QP

        xt = np.array([0.0, 1.0, 2.0, 3.0, 4.0])
        yt = np.array([0.0, 1.0, 1.5, 0.5, 1.0])

        # Specify the design space using the DesignSpace
        # class and various available variable types
        design_space = DesignSpace(
            [
                IntegerVariable(0, 4),
            ]
        )
        sm = MixedIntegerSurrogateModel(design_space=design_space, surrogate=QP())
        sm.set_training_values(xt, yt)
        sm.train()

        num = 100
        x = np.linspace(0.0, 4.0, num)
        y = sm.predict_values(x)

        plt.plot(xt, yt, "o")
        plt.plot(x, y)
        plt.xlabel("x")
        plt.ylabel("y")
        plt.legend(["Training data", "Prediction"])
        plt.show()

    def run_mixed_integer_context_example(self):
        import matplotlib.pyplot as plt

        from smt.applications.mixed_integer import MixedIntegerContext
        from smt.design_space import (
            CategoricalVariable,
            DesignSpace,
            FloatVariable,
            IntegerVariable,
        )
        from smt.surrogate_models import KRG

        design_space = DesignSpace(
            [
                IntegerVariable(0, 5),
                FloatVariable(0.0, 4.0),
                CategoricalVariable(["blue", "red", "green", "yellow"]),
            ]
        )

        def ftest(x):
            return (x[:, 0] * x[:, 0] + x[:, 1] * x[:, 1]) * (x[:, 2] + 1)

        # Helper class for creating surrogate models
        mi_context = MixedIntegerContext(design_space)

        # DOE for training
        sampler = mi_context.build_sampling_method()

        num = mi_context.get_unfolded_dimension() * 5
        print("DOE point nb = {}".format(num))
        xt = sampler(num)
        yt = ftest(xt)

        # Surrogate
        sm = mi_context.build_kriging_model(KRG(hyper_opt="Cobyla"))
        sm.set_training_values(xt, yt)
        sm.train()

        # DOE for validation
        xv = sampler(50)
        yv = ftest(xv)
        yp = sm.predict_values(xv)

        plt.plot(yv, yv)
        plt.plot(yv, yp, "o")
        plt.xlabel("actual")
        plt.ylabel("prediction")

        plt.show()

    def test_hierarchical_variables_Goldstein(self):
        problem = HierarchicalGoldstein()
        ds = problem.design_space
        self.assertIsInstance(ds, DesignSpace)
        self.assertEqual(ds.n_dv, 11)

        x = np.array(
            [
                [0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1],
                [1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1],
                [2, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1],
                [3, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1],
            ]
        )
        y = problem(x)
        self.assertTrue(
            np.linalg.norm(
                y - np.array([50.75285716, 56.62074043, 50.97693309, 56.29235443])
            )
            < 1e-8
        )
        self.assertTrue(
            np.linalg.norm(
                problem.eval_is_acting.astype(int)
                - np.array(
                    [
                        [1, 1, 1, 1, 0, 0, 1, 1, 1, 1, 1],
                        [1, 1, 1, 1, 1, 0, 1, 0, 1, 1, 1],
                        [1, 1, 1, 1, 0, 1, 1, 1, 0, 1, 1],
                        [1, 1, 1, 1, 1, 1, 1, 0, 0, 1, 1],
                    ]
                )
            )
            < 1e-8
        )
        self.assertTrue(
            np.linalg.norm(
                problem.eval_x
                - np.array(
                    [
                        [0, 1, 1, 1, 50, 50, 1, 1, 1, 1, 1],
                        [1, 1, 1, 1, 1, 50, 1, 0, 1, 1, 1],
                        [2, 1, 1, 1, 50, 1, 1, 1, 0, 1, 1],
                        [3, 1, 1, 1, 1, 1, 1, 0, 0, 1, 1],
                    ]
                )
            )
            < 1e-8
        )
        self.assertTrue(np.linalg.norm(y - problem(problem.eval_x)) < 1e-8)

        n_doe = 15
        ds.seed = 42
        samp = MixedIntegerSamplingMethod(
            LHS, ds, criterion="ese", random_state=ds.seed
        )
        Xt, is_acting = samp(n_doe, return_is_acting=True)

        Yt = problem(Xt)

        sm = MixedIntegerKrigingModel(
            surrogate=KRG(
                design_space=ds,
                categorical_kernel=MixIntKernelType.HOMO_HSPHERE,
                hierarchical_kernel=MixHrcKernelType.ARC_KERNEL,
                hyper_opt="Cobyla",
                theta0=[1e-2],
                corr="abs_exp",
                n_start=10,
            ),
        )
        sm.set_training_values(Xt, Yt, is_acting=is_acting)
        sm.train()
        y_s = sm.predict_values(Xt)[:, 0]
        pred_RMSE = np.linalg.norm(y_s - Yt) / len(Yt)

        y_sv = sm.predict_variances(Xt)[:, 0]
        var_RMSE = np.linalg.norm(y_sv) / len(Yt)
        self.assertLess(pred_RMSE, 1e-7)
        self.assertLess(var_RMSE, 1e-7)
        self.assertTrue(
            np.linalg.norm(
                sm.predict_values(
                    np.array(
                        [
                            [0.0, 1.0, 64.0, 4.0, 56.0, 37.0, 35.0, 1.0, 2.0, 1.0, 1.0],
                            [1.0, 0.0, 31.0, 92.0, 24.0, 3.0, 17.0, 1.0, 2.0, 1.0, 1.0],
                            [2.0, 1.0, 28.0, 60.0, 77.0, 66.0, 9.0, 0.0, 1.0, 1.0, 1.0],
                            [
                                3.0,
                                1.0,
                                50.0,
                                40.0,
                                99.0,
                                35.0,
                                51.0,
                                2.0,
                                1.0,
                                1.0,
                                2.0,
                            ],
                        ]
                    )
                )[:, 0]
                - sm.predict_values(
                    np.array(
                        [
                            [0.0, 1.0, 64.0, 4.0, 6.0, 7.0, 35.0, 1.0, 2.0, 1.0, 1.0],
                            [
                                1.0,
                                0.0,
                                31.0,
                                92.0,
                                24.0,
                                30.0,
                                17.0,
                                0.0,
                                2.0,
                                1.0,
                                1.0,
                            ],
                            [2.0, 1.0, 28.0, 60.0, 7.0, 66.0, 9.0, 0.0, 2.0, 1.0, 1.0],
                            [
                                3.0,
                                1.0,
                                50.0,
                                40.0,
                                99.0,
                                35.0,
                                51.0,
                                0.0,
                                0.0,
                                1.0,
                                2.0,
                            ],
                        ]
                    )
                )[:, 0]
            )
            < 1e-8
        )
        self.assertTrue(
            np.linalg.norm(
                sm.predict_values(
                    np.array(
                        [[1.0, 0.0, 31.0, 92.0, 24.0, 3.0, 17.0, 1.0, 2.0, 1.0, 1.0]]
                    )
                )
                - sm.predict_values(
                    np.array(
                        [[1.0, 0.0, 31.0, 92.0, 24.0, 3.0, 17.0, 1.0, 1.0, 1.0, 1.0]]
                    )
                )
            )
            > 1e-8
        )

    def run_mixed_discrete_design_space_example(self):
        import numpy as np

        from smt.applications.mixed_integer import MixedIntegerSamplingMethod
        from smt.design_space import (
            CategoricalVariable,
            DesignSpace,
            FloatVariable,
            IntegerVariable,
            OrdinalVariable,
        )
        from smt.sampling_methods import LHS

        ds = DesignSpace(
            [
                CategoricalVariable(
                    ["A", "B"]
                ),  # x0 categorical: A or B; order is not relevant
                OrdinalVariable(
                    ["C", "D", "E"]
                ),  # x1 ordinal: C, D or E; order is relevant
                IntegerVariable(
                    0, 2
                ),  # x2 integer between 0 and 2 (inclusive): 0, 1, 2
                FloatVariable(0, 1),  # c3 continuous between 0 and 1
            ]
        )

        # Sample the design space
        # Note: is_acting_sampled specifies for each design variable whether it is acting or not
        ds.seed = 42
        samp = MixedIntegerSamplingMethod(
            LHS, ds, criterion="ese", random_state=ds.seed
        )
        x_sampled, is_acting_sampled = samp(100, return_is_acting=True)

        # Correct design vectors: round discrete variables, correct hierarchical variables
        x_corr, is_acting = ds.correct_get_acting(
            np.array(
                [
                    [0, 0, 2, 0.25],
                    [0, 2, 1, 0.75],
                ]
            )
        )
        print(is_acting)

    def run_hierarchical_design_space_example(self):
        import numpy as np
        from smt_design_space_ext import ConfigSpaceDesignSpaceImpl

        from smt.applications.mixed_integer import (
            MixedIntegerKrigingModel,
            MixedIntegerSamplingMethod,
        )
        from smt.design_space import (
            CategoricalVariable,
            FloatVariable,
            IntegerVariable,
            OrdinalVariable,
        )
        from smt.sampling_methods import LHS
        from smt.surrogate_models import KRG, MixHrcKernelType, MixIntKernelType

        ds = ConfigSpaceDesignSpaceImpl(
            [
                CategoricalVariable(
                    ["A", "B"]
                ),  # x0 categorical: A or B; order is not relevant
                OrdinalVariable(
                    ["C", "D", "E"]
                ),  # x1 ordinal: C, D or E; order is relevant
                IntegerVariable(
                    0, 2
                ),  # x2 integer between 0 and 2 (inclusive): 0, 1, 2
                FloatVariable(0, 1),  # c3 continuous between 0 and 1
            ]
        )

        # Declare that x1 is acting if x0 == A
        ds.declare_decreed_var(decreed_var=1, meta_var=0, meta_value="A")

        # Nested hierarchy is possible: activate x2 if x1 == C or D
        # Note: only if ConfigSpace is installed! pip install smt[cs]
        ds.declare_decreed_var(decreed_var=2, meta_var=1, meta_value=["C", "D"])

        # It is also possible to explicitly forbid two values from occurring simultaneously
        # Note: only if ConfigSpace is installed! pip install smt[cs]
        ds.add_value_constraint(
            var1=0, value1="A", var2=2, value2=[0, 1]
        )  # Forbid x0 == A && x2 == 0 or 1

        # For quantitative variables, it is possible to specify order relation
        ds.add_value_constraint(
            var1=2, value1="<", var2=3, value2=">"
        )  # Prevent x2 < x3

        # Sample the design space
        # Note: is_acting_sampled specifies for each design variable whether it is acting or not
        ds.seed = 42
        samp = MixedIntegerSamplingMethod(
            LHS, ds, criterion="ese", random_state=ds.seed
        )
        Xt, is_acting_sampled = samp(100, return_is_acting=True)

        rng = np.random.default_rng(42)
        Yt = 4 * rng.random(100) - 2 + Xt[:, 0] + Xt[:, 1] - Xt[:, 2] - Xt[:, 3]
        # Correct design vectors: round discrete variables, correct hierarchical variables
        x_corr, is_acting = ds.correct_get_acting(
            np.array(
                [
                    [0, 0, 2, 0.25],
                    [0, 2, 1, 0.75],
                    [1, 2, 1, 0.66],
                ]
            )
        )

        # Observe the hierarchical behavior:
        assert np.all(
            is_acting
            == np.array(
                [
                    [True, True, True, True],
                    [
                        True,
                        True,
                        False,
                        True,
                    ],  # x2 is not acting if x1 != C or D (0 or 1)
                    [
                        True,
                        False,
                        False,
                        True,
                    ],  # x1 is not acting if x0 != A, and x2 is not acting because x1 is not acting
                ]
            )
        )
        assert np.all(
            x_corr
            == np.array(
                [
                    [0, 0, 2, 0.25],
                    [0, 2, 0, 0.75],
                    # x2 is not acting, so it is corrected ("imputed") to its non-acting value (0 for discrete vars)
                    [1, 0, 0, 0.66],  # x1 and x2 are imputed
                ]
            )
        )

        sm = MixedIntegerKrigingModel(
            surrogate=KRG(
                design_space=ds,
                categorical_kernel=MixIntKernelType.HOMO_HSPHERE,
                hierarchical_kernel=MixHrcKernelType.ALG_KERNEL,
                theta0=[1e-2],
                hyper_opt="Cobyla",
                corr="abs_exp",
                n_start=5,
            ),
        )
        sm.set_training_values(Xt, Yt)
        sm.train()
        y_s = sm.predict_values(Xt)[:, 0]
        pred_RMSE = np.linalg.norm(y_s - Yt) / len(Yt)

        y_sv = sm.predict_variances(Xt)[:, 0]
        _var_RMSE = np.linalg.norm(y_sv) / len(Yt)
        assert pred_RMSE < 1e-7
        print("Pred_RMSE", pred_RMSE)

        self._sm = sm  # to be ignored: just used for automated test

    @unittest.skip(
        "Use smt design space extension capability, this test should be moved in smt_design_space_ext"
    )
    def test_hierarchical_design_space_example(self):
        self.run_hierarchical_design_space_example()
        np.testing.assert_almost_equal(
            self._sm.predict_values(
                np.array(
                    [
                        [0, 2, 1, 0.75],
                        [1, 2, 1, 0.66],
                        [0, 2, 1, 0.75],
                    ]
                )
            )[:, 0],
            self._sm.predict_values(
                np.array(
                    [
                        [0, 2, 2, 0.75],
                        [1, 1, 2, 0.66],
                        [0, 2, 0, 0.75],
                    ]
                )
            )[:, 0],
        )
        self.assertTrue(
            np.linalg.norm(
                self._sm.predict_values(np.array([[0, 0, 2, 0.25]]))
                - self._sm.predict_values(np.array([[0, 0, 2, 0.8]]))
            )
            > 1e-8
        )

    @unittest.skip(
        "Use smt design space extension capability, this test should be moved in smt_design_space_ext"
    )
    def test_hierarchical_design_space_example_all_categorical_decreed(self):
        ds = DesignSpace(
            [
                CategoricalVariable(
                    ["A", "B"]
                ),  # x0 categorical: A or B; order is not relevant
                CategoricalVariable(
                    ["C", "D", "E"]
                ),  # x1 ordinal: C, D or E; order is relevant
                CategoricalVariable(
                    ["tata", "tutu", "toto"]
                ),  # x2 integer between 0 and 2 (inclusive): 0, 1, 2
                FloatVariable(0, 1),  # c3 continuous between 0 and 1
            ]
        )

        # Declare that x1 is acting if x0 == A
        ds.declare_decreed_var(decreed_var=1, meta_var=0, meta_value="A")

        # Nested hierarchy is possible: activate x2 if x1 == C or D
        # Note: only if ConfigSpace is installed! pip install smt[cs]
        ds.declare_decreed_var(decreed_var=2, meta_var=1, meta_value=["C", "D"])

        # It is also possible to explicitly forbid two values from occurring simultaneously
        # Note: only if ConfigSpace is installed! pip install smt[cs]
        # ds.add_value_constraint(
        #    var1=0, value1="A", var2=2, value2=["tata","tutu"]
        # )  # Forbid x0 == A && x2 == 0 or 1

        # Sample the design space
        # Note: is_acting_sampled specifies for each design variable whether it is acting or not
        ds.seed = 42
        samp = MixedIntegerSamplingMethod(
            LHS, ds, criterion="ese", random_state=ds.seed
        )
        Xt, is_acting_sampled = samp(100, return_is_acting=True)

        rng = np.random.default_rng(42)
        Yt = 4 * rng.random(100) - 2 + Xt[:, 0] + Xt[:, 1] - Xt[:, 2] - Xt[:, 3]
        # Correct design vectors: round discrete variables, correct hierarchical variables
        x_corr, is_acting = ds.correct_get_acting(
            np.array(
                [
                    [0, 0, 2, 0.25],
                    [0, 2, 1, 0.75],
                    [1, 2, 1, 0.66],
                ]
            )
        )

        # Observe the hierarchical behavior:
        np.testing.assert_array_equal(
            is_acting,
            np.array(
                [
                    [True, True, True, True],
                    [
                        True,
                        True,
                        False,
                        True,
                    ],  # x2 is not acting if x1 != C or D (0 or 1)
                    [
                        True,
                        False,
                        False,
                        True,
                    ],  # x1 is not acting if x0 != A, and x2 is not acting because x1 is not acting
                ]
            ),
        )

        np.testing.assert_array_equal(
            x_corr,
            np.array(
                [
                    [0, 0, 2, 0.25],
                    [0, 2, 0, 0.75],
                    # x2 is not acting, so it is corrected ("imputed") to its non-acting value (0 for discrete vars)
                    [1, 0, 0, 0.66],  # x1 and x2 are imputed
                ]
            ),
        )

        for mixint_kernel in [
            MixIntKernelType.CONT_RELAX,
            MixIntKernelType.GOWER,
            MixIntKernelType.COMPOUND_SYMMETRY,
            MixIntKernelType.EXP_HOMO_HSPHERE,
            MixIntKernelType.HOMO_HSPHERE,
        ]:
            sm = MixedIntegerKrigingModel(
                surrogate=KRG(
                    design_space=ds,
                    categorical_kernel=mixint_kernel,
                    hierarchical_kernel=MixHrcKernelType.ALG_KERNEL,
                    theta0=[1e-2],
                    hyper_opt="Cobyla",
                    corr="abs_exp",
                    n_start=5,
                ),
            )
            sm.set_training_values(Xt, Yt)
            sm.train()
            y_s = sm.predict_values(Xt)[:, 0]
            _pred_RMSE = np.linalg.norm(y_s - Yt) / len(Yt)
            self.assertLess(_pred_RMSE, 1e-6)
            y_sv = sm.predict_variances(Xt)[:, 0]
            _var_RMSE = np.linalg.norm(y_sv) / len(Yt)
            self.assertLess(_var_RMSE, 1e-6)
            np.testing.assert_almost_equal(
                sm.predict_values(
                    np.array(
                        [
                            [0, 2, 1, 0.75],
                            [1, 2, 1, 0.66],
                            [0, 2, 1, 0.75],
                        ]
                    )
                )[:, 0],
                sm.predict_values(
                    np.array(
                        [
                            [0, 2, 2, 0.75],
                            [1, 1, 2, 0.66],
                            [0, 2, 0, 0.75],
                        ]
                    )
                )[:, 0],
            )
            self.assertTrue(
                np.linalg.norm(
                    sm.predict_values(np.array([[0, 0, 2, 0.25]]))
                    - sm.predict_values(np.array([[0, 0, 2, 0.8]]))
                )
                > 1e-8
            )

    def run_hierarchical_variables_Goldstein(self):
        import numpy as np

        from smt.applications.mixed_integer import (
            MixedIntegerKrigingModel,
            MixedIntegerSamplingMethod,
        )
        from smt.design_space import (
            CategoricalVariable,
            DesignSpace,
            FloatVariable,
            IntegerVariable,
        )
        from smt.sampling_methods import LHS
        from smt.surrogate_models import KRG, MixHrcKernelType, MixIntKernelType

        def f_hv(X):
            import numpy as np

            def H(x1, x2, x3, x4, z3, z4, x5, cos_term):
                import numpy as np

                h = (
                    53.3108
                    + 0.184901 * x1
                    - 5.02914 * x1**3 * 10 ** (-6)
                    + 7.72522 * x1**z3 * 10 ** (-8)
                    - 0.0870775 * x2
                    - 0.106959 * x3
                    + 7.98772 * x3**z4 * 10 ** (-6)
                    + 0.00242482 * x4
                    + 1.32851 * x4**3 * 10 ** (-6)
                    - 0.00146393 * x1 * x2
                    - 0.00301588 * x1 * x3
                    - 0.00272291 * x1 * x4
                    + 0.0017004 * x2 * x3
                    + 0.0038428 * x2 * x4
                    - 0.000198969 * x3 * x4
                    + 1.86025 * x1 * x2 * x3 * 10 ** (-5)
                    - 1.88719 * x1 * x2 * x4 * 10 ** (-6)
                    + 2.50923 * x1 * x3 * x4 * 10 ** (-5)
                    - 5.62199 * x2 * x3 * x4 * 10 ** (-5)
                )
                if cos_term:
                    h += 5.0 * np.cos(2.0 * np.pi * (x5 / 100.0)) - 2.0
                return h

            def f1(x1, x2, z1, z2, z3, z4, x5, cos_term):
                c1 = z2 == 0
                c2 = z2 == 1
                c3 = z2 == 2

                c4 = z3 == 0
                c5 = z3 == 1
                c6 = z3 == 2

                y = (
                    c4
                    * (
                        c1 * H(x1, x2, 20, 20, z3, z4, x5, cos_term)
                        + c2 * H(x1, x2, 50, 20, z3, z4, x5, cos_term)
                        + c3 * H(x1, x2, 80, 20, z3, z4, x5, cos_term)
                    )
                    + c5
                    * (
                        c1 * H(x1, x2, 20, 50, z3, z4, x5, cos_term)
                        + c2 * H(x1, x2, 50, 50, z3, z4, x5, cos_term)
                        + c3 * H(x1, x2, 80, 50, z3, z4, x5, cos_term)
                    )
                    + c6
                    * (
                        c1 * H(x1, x2, 20, 80, z3, z4, x5, cos_term)
                        + c2 * H(x1, x2, 50, 80, z3, z4, x5, cos_term)
                        + c3 * H(x1, x2, 80, 80, z3, z4, x5, cos_term)
                    )
                )
                return y

            def f2(x1, x2, x3, z2, z3, z4, x5, cos_term):
                c1 = z2 == 0
                c2 = z2 == 1
                c3 = z2 == 2

                y = (
                    c1 * H(x1, x2, x3, 20, z3, z4, x5, cos_term)
                    + c2 * H(x1, x2, x3, 50, z3, z4, x5, cos_term)
                    + c3 * H(x1, x2, x3, 80, z3, z4, x5, cos_term)
                )
                return y

            def f3(x1, x2, x4, z1, z3, z4, x5, cos_term):
                c1 = z1 == 0
                c2 = z1 == 1
                c3 = z1 == 2

                y = (
                    c1 * H(x1, x2, 20, x4, z3, z4, x5, cos_term)
                    + c2 * H(x1, x2, 50, x4, z3, z4, x5, cos_term)
                    + c3 * H(x1, x2, 80, x4, z3, z4, x5, cos_term)
                )
                return y

            y = []
            for x in X:
                if x[0] == 0:
                    y.append(
                        f1(x[2], x[3], x[7], x[8], x[9], x[10], x[6], cos_term=x[1])
                    )
                elif x[0] == 1:
                    y.append(
                        f2(x[2], x[3], x[4], x[8], x[9], x[10], x[6], cos_term=x[1])
                    )
                elif x[0] == 2:
                    y.append(
                        f3(x[2], x[3], x[5], x[7], x[9], x[10], x[6], cos_term=x[1])
                    )
                elif x[0] == 3:
                    y.append(
                        H(x[2], x[3], x[4], x[5], x[9], x[10], x[6], cos_term=x[1])
                    )
            return np.array(y)

        design_space = DesignSpace(
            [
                CategoricalVariable(values=[0, 1, 2, 3]),  # meta
                IntegerVariable(0, 1),  # x1
                FloatVariable(0, 100),  # x2
                FloatVariable(0, 100),
                FloatVariable(0, 100),
                FloatVariable(0, 100),
                FloatVariable(0, 100),
                IntegerVariable(0, 2),  # x7
                IntegerVariable(0, 2),
                IntegerVariable(0, 2),
                IntegerVariable(0, 2),
            ]
        )

        # x4 is acting if meta == 1, 3
        design_space.declare_decreed_var(decreed_var=4, meta_var=0, meta_value=[1, 3])
        # x5 is acting if meta == 2, 3
        design_space.declare_decreed_var(decreed_var=5, meta_var=0, meta_value=[2, 3])
        # x7 is acting if meta == 0, 2
        design_space.declare_decreed_var(decreed_var=7, meta_var=0, meta_value=[0, 2])
        # x8 is acting if meta == 0, 1
        design_space.declare_decreed_var(decreed_var=8, meta_var=0, meta_value=[0, 1])

        # Sample from the design spaces, correctly considering hierarchy
        n_doe = 15
        design_space.seed = 42
        samp = MixedIntegerSamplingMethod(
            LHS, design_space, criterion="ese", random_state=design_space.seed
        )
        Xt, Xt_is_acting = samp(n_doe, return_is_acting=True)

        Yt = f_hv(Xt)

        sm = MixedIntegerKrigingModel(
            surrogate=KRG(
                design_space=design_space,
                categorical_kernel=MixIntKernelType.HOMO_HSPHERE,
                hierarchical_kernel=MixHrcKernelType.ALG_KERNEL,  # ALG or ARC
                theta0=[1e-2],
                hyper_opt="Cobyla",
                corr="abs_exp",
                n_start=5,
            ),
        )
        sm.set_training_values(Xt, Yt, is_acting=Xt_is_acting)
        sm.train()
        y_s = sm.predict_values(Xt)[:, 0]
        _pred_RMSE = np.linalg.norm(y_s - Yt) / len(Yt)

        y_sv = sm.predict_variances(Xt)[:, 0]
        _var_RMSE = np.linalg.norm(y_sv) / len(Yt)

    @unittest.skipIf(int(os.getenv("RUN_SLOW_TESTS", 0)) < 1, "too slow")
    def test_hierarchical_variables_NN(self):
        problem = HierarchicalNeuralNetwork()
        ds = problem.design_space
        self.assertEqual(ds.n_dv, 8)
        n_doe = 100

        ds_sampling = DesignSpace(ds.design_variables[1:])
        sampling = MixedIntegerSamplingMethod(
            LHS, ds_sampling, criterion="ese", random_state=42
        )
        x_cont = sampling(3 * n_doe)

        xdoe1 = np.zeros((n_doe, 8))
        x_cont2 = x_cont[:n_doe, :5]
        xdoe1[:, 0] = np.zeros(n_doe)
        xdoe1[:, 1:6] = x_cont2
        ydoe1 = problem(xdoe1)

        xdoe1 = np.zeros((n_doe, 8))
        xdoe1[:, 0] = np.zeros(n_doe)
        xdoe1[:, 1:6] = x_cont2

        xdoe2 = np.zeros((n_doe, 8))
        x_cont2 = x_cont[n_doe : 2 * n_doe, :6]
        xdoe2[:, 0] = np.ones(n_doe)
        xdoe2[:, 1:7] = x_cont2
        ydoe2 = problem(xdoe2)

        xdoe2 = np.zeros((n_doe, 8))
        xdoe2[:, 0] = np.ones(n_doe)
        xdoe2[:, 1:7] = x_cont2

        xdoe3 = np.zeros((n_doe, 8))
        xdoe3[:, 0] = 2 * np.ones(n_doe)
        xdoe3[:, 1:] = x_cont[2 * n_doe :, :]
        ydoe3 = problem(xdoe3)

        Xt = np.concatenate((xdoe1, xdoe2, xdoe3), axis=0)
        Yt = np.concatenate((ydoe1, ydoe2, ydoe3), axis=0)
        sm = MixedIntegerKrigingModel(
            surrogate=KRG(
                design_space=problem.design_space,
                categorical_kernel=MixIntKernelType.HOMO_HSPHERE,
                hierarchical_kernel=MixHrcKernelType.ALG_KERNEL,
                theta0=[1e-2],
                hyper_opt="Cobyla",
                corr="abs_exp",
                n_start=5,
            ),
        )
        sm.set_training_values(Xt, Yt)
        sm.train()
        y_s = sm.predict_values(Xt)[:, 0]
        pred_RMSE = np.linalg.norm(y_s - Yt) / len(Yt)

        y_sv = sm.predict_variances(Xt)[:, 0]
        var_RMSE = np.linalg.norm(y_sv) / len(Yt)
        self.assertLess(pred_RMSE, 1e-7)
        self.assertLess(var_RMSE, 1e-7)
        np.testing.assert_almost_equal(
            sm.predict_values(
                np.array(
                    [
                        [0, -1, -2, 0, 0, 2, 0, 0],
                        [1, -1, -2, 1, 1, 2, 1, 0],
                        [2, -1, -2, 2, 2, 2, 1, -2],
                    ]
                )
            )[:, 0],
            sm.predict_values(
                np.array(
                    [
                        [0, -1, -2, 0, 0, 2, 10, 10],
                        [1, -1, -2, 1, 1, 2, 1, 10],
                        [2, -1, -2, 2, 2, 2, 1, -2],
                    ]
                )
            )[:, 0],
        )
        self.assertTrue(
            np.linalg.norm(
                sm.predict_values(np.array([[0, -1, -2, 0, 0, 2, 0, 0]]))
                - sm.predict_values(np.array([[0, -1, -2, 0, 0, 12, 10, 10]]))
            )
            > 1e-8
        )

    def test_mixed_gower_2D(self):
        xt = np.array([[0, 5], [2, -1], [4, 0.5]])
        yt = np.array([[0.0], [1.0], [1.5]])
        design_space = DesignSpace(
            [
                CategoricalVariable(["0.0", "1.0", " 2.0", "3.0", "4.0"]),
                FloatVariable(-5, 5),
            ]
        )

        # Surrogate
        sm = MixedIntegerKrigingModel(
            surrogate=KRG(
                design_space=design_space,
                theta0=[1e-2],
                hyper_opt="Cobyla",
                corr="abs_exp",
                categorical_kernel=MixIntKernelType.GOWER,
            ),
        )
        sm.set_training_values(xt, yt)
        sm.train()

        # DOE for validation
        x = np.linspace(0, 4, 5)
        x2 = np.linspace(-5, 5, 21)
        x1 = []
        for element in itertools.product(x, x2):
            x1.append(np.array(element))
        x_pred = np.array(x1)

        y = sm.predict_values(x_pred)
        yvar = sm.predict_variances(x_pred)

        # prediction are correct on known points
        self.assertAlmostEqual(y[20, 0], 0)
        self.assertAlmostEqual(y[50, 0], 1)
        self.assertAlmostEqual(y[95, 0], 1.5)
        self.assertTrue(np.abs(np.sum(np.array([y[20], y[50], y[95]]) - yt)) < 1e-6)
        self.assertTrue(np.abs(np.sum(np.array([yvar[20], yvar[50], yvar[95]]))) < 1e-6)

        self.assertEqual(np.shape(y), (105, 1))

    def test_mixed_cs_2D(self):
        xt = np.array([[0, 5], [2, -1], [4, 0.5]])
        yt = np.array([[0.0], [1.0], [1.5]])
        design_space = DesignSpace(
            [
                CategoricalVariable(["0.0", "1.0", " 2.0", "3.0", "4.0"]),
                FloatVariable(-5, 5),
            ]
        )

        # Surrogate
        sm = MixedIntegerKrigingModel(
            surrogate=KRG(
                design_space=design_space,
                theta0=[1e-2],
                hyper_opt="Cobyla",
                corr="abs_exp",
                categorical_kernel=MixIntKernelType.COMPOUND_SYMMETRY,
            ),
        )
        sm.set_training_values(xt, yt)
        sm.train()

        # DOE for validation
        x = np.linspace(0, 4, 5)
        x2 = np.linspace(-5, 5, 21)
        x1 = []
        for element in itertools.product(x, x2):
            x1.append(np.array(element))
        x_pred = np.array(x1)

        y = sm.predict_values(x_pred)
        yvar = sm.predict_variances(x_pred)

        # prediction are correct on known points
        self.assertAlmostEqual(y[20, 0], 0)
        self.assertAlmostEqual(y[50, 0], 1)
        self.assertAlmostEqual(y[95, 0], 1.5)
        self.assertTrue(np.abs(np.sum(np.array([y[20], y[50], y[95]]) - yt)) < 1e-6)
        self.assertTrue(np.abs(np.sum(np.array([yvar[20], yvar[50], yvar[95]]))) < 1e-6)

        self.assertEqual(np.shape(y), (105, 1))

    def test_mixed_CR_2D(self):
        xt = np.array([[0, 5], [2, -1], [4, 0.5]])
        yt = np.array([[0.0], [1.0], [1.5]])
        design_space = DesignSpace(
            [
                CategoricalVariable(["0.0", "1.0", " 2.0", "3.0", "4.0"]),
                FloatVariable(-5, 5),
            ]
        )

        # Surrogate
        sm = MixedIntegerKrigingModel(
            surrogate=KRG(
                design_space=design_space,
                theta0=[1e-2],
                hyper_opt="Cobyla",
                corr="abs_exp",
                categorical_kernel=MixIntKernelType.GOWER,
            ),
        )
        sm.set_training_values(xt, yt)
        sm.train()

        # DOE for validation
        x = np.linspace(0, 4, 5)
        x2 = np.linspace(-5, 5, 21)
        x1 = []
        for element in itertools.product(x, x2):
            x1.append(np.array(element))
        x_pred = np.array(x1)

        y = sm.predict_values(x_pred)
        yvar = sm.predict_variances(x_pred)

        # prediction are correct on known points
        self.assertAlmostEqual(y[20, 0], 0)
        self.assertAlmostEqual(y[50, 0], 1)
        self.assertAlmostEqual(y[95, 0], 1.5)
        self.assertTrue(np.abs(np.sum(np.array([y[20], y[50], y[95]]) - yt)) < 1e-6)
        self.assertTrue(np.abs(np.sum(np.array([yvar[20], yvar[50], yvar[95]]))) < 1e-6)

        self.assertEqual(np.shape(y), (105, 1))

    def test_mixed_CR_PLS_noisy_2D(self):
        xt = np.array([[0, 5], [2, -1], [4, 0.5]])
        yt = np.array([[0.0], [1.0], [1.5]])
        design_space = DesignSpace(
            [
                CategoricalVariable(["0.0", "1.0", " 2.0", "3.0", "4.0"]),
                FloatVariable(-5, 5),
            ]
        )
        # Surrogate
        sm = MixedIntegerKrigingModel(
            surrogate=KPLS(
                n_comp=1,
                eval_noise=True,
                design_space=design_space,
                theta0=[1e-2],
                categorical_kernel=MixIntKernelType.CONT_RELAX,
                hyper_opt="Cobyla",
                corr="abs_exp",
            ),
        )
        sm.set_training_values(xt, yt)
        sm.train()

        # DOE for validation
        x = np.linspace(0, 4, 5)
        x2 = np.linspace(-5, 5, 21)
        x1 = []
        for element in itertools.product(x, x2):
            x1.append(np.array(element))
        x_pred = np.array(x1)

        y = sm.predict_values(x_pred)
        yvar = sm.predict_variances(x_pred)

        # prediction are correct on known points
        self.assertTrue(np.abs(np.sum(np.array([y[20], y[50], y[95]]) - yt)) < 1e-6)
        self.assertTrue(np.abs(np.sum(np.array([yvar[20], yvar[50], yvar[95]]))) < 1e-6)

        self.assertEqual(np.shape(y), (105, 1))

    def test_mixed_CR_PLS_2D(self):
        xt = np.array([[0, 5], [2, -1], [4, 0.5]])
        yt = np.array([[0.0], [1.0], [1.5]])
        design_space = DesignSpace(
            [
                CategoricalVariable(["0.0", "1.0", " 2.0", "3.0", "4.0"]),
                FloatVariable(-5, 5),
            ]
        )
        # Surrogate
        sm = MixedIntegerKrigingModel(
            surrogate=KPLS(
                n_comp=1,
                eval_noise=False,
                design_space=design_space,
                theta0=[1e-2],
                categorical_kernel=MixIntKernelType.CONT_RELAX,
                hyper_opt="Cobyla",
                corr="abs_exp",
            ),
        )
        sm.set_training_values(xt, yt)
        sm.train()

        # DOE for validation
        x = np.linspace(0, 4, 5)
        x2 = np.linspace(-5, 5, 21)
        x1 = []
        for element in itertools.product(x, x2):
            x1.append(np.array(element))
        x_pred = np.array(x1)

        y = sm.predict_values(x_pred)
        yvar = sm.predict_variances(x_pred)

        # prediction are correct on known points
        self.assertTrue(np.abs(np.sum(np.array([y[20], y[50], y[95]]) - yt)) < 1e-6)
        self.assertTrue(np.abs(np.sum(np.array([yvar[20], yvar[50], yvar[95]]))) < 1e-6)

        self.assertEqual(np.shape(y), (105, 1))

    def test_mixed_CR_KPLSK_2D(self):
        xt = np.array([[0, 5], [2, -1], [4, 0.5]])
        yt = np.array([[0.0], [1.0], [1.5]])
        design_space = DesignSpace(
            [
                CategoricalVariable(["0.0", "1.0", " 2.0", "3.0", "4.0"]),
                FloatVariable(-5, 5),
            ]
        )
        # Surrogate
        sm = MixedIntegerKrigingModel(
            surrogate=KPLSK(
                n_comp=1,
                eval_noise=False,
                design_space=design_space,
                theta0=[1e-2],
                categorical_kernel=MixIntKernelType.CONT_RELAX,
                hyper_opt="Cobyla",
                corr="abs_exp",
            ),
        )
        sm.set_training_values(xt, yt)
        sm.train()

        # DOE for validation
        x = np.linspace(0, 4, 5)
        x2 = np.linspace(-5, 5, 21)
        x1 = []
        for element in itertools.product(x, x2):
            x1.append(np.array(element))
        x_pred = np.array(x1)

        y = sm.predict_values(x_pred)
        yvar = sm.predict_variances(x_pred)

        # prediction are correct on known points
        self.assertTrue(np.abs(np.sum(np.array([y[20], y[50], y[95]]) - yt)) < 1e-6)
        self.assertTrue(np.abs(np.sum(np.array([yvar[20], yvar[50], yvar[95]]))) < 1e-6)

        self.assertEqual(np.shape(y), (105, 1))

    def test_mixed_GD_KPLSK_2D(self):
        xt = np.array([[0, 5], [2, -1], [4, 0.5]])
        yt = np.array([[0.0], [1.0], [1.5]])
        design_space = DesignSpace(
            [
                CategoricalVariable(["0.0", "1.0", " 2.0", "3.0", "4.0"]),
                FloatVariable(-5, 5),
            ]
        )
        # Surrogate
        sm = MixedIntegerKrigingModel(
            surrogate=KPLSK(
                n_comp=1,
                eval_noise=False,
                design_space=design_space,
                theta0=[1e-2],
                categorical_kernel=MixIntKernelType.GOWER,
                hyper_opt="Cobyla",
                corr="abs_exp",
            ),
        )
        sm.set_training_values(xt, yt)
        sm.train()

        # DOE for validation
        x = np.linspace(0, 4, 5)
        x2 = np.linspace(-5, 5, 21)
        x1 = []
        for element in itertools.product(x, x2):
            x1.append(np.array(element))
        x_pred = np.array(x1)

        y = sm.predict_values(x_pred)
        yvar = sm.predict_variances(x_pred)

        # prediction are correct on known points
        self.assertTrue(np.abs(np.sum(np.array([y[20], y[50], y[95]]) - yt)) < 1e-6)
        self.assertTrue(np.abs(np.sum(np.array([yvar[20], yvar[50], yvar[95]]))) < 1e-6)

        self.assertEqual(np.shape(y), (105, 1))

    def test_mixed_homo_gaussian_2D(self):
        xt = np.array([[0, 5], [2, -1], [4, 0.5]])
        yt = np.array([[0.0], [1.0], [1.5]])
        design_space = DesignSpace(
            [
                CategoricalVariable(["0.0", "1.0", " 2.0", "3.0", "4.0"]),
                FloatVariable(-5, 5),
            ]
        )

        # Surrogate
        sm = MixedIntegerKrigingModel(
            surrogate=KRG(
                design_space=design_space,
                theta0=[1e-2],
                corr="abs_exp",
                hyper_opt="Cobyla",
                categorical_kernel=MixIntKernelType.EXP_HOMO_HSPHERE,
            ),
        )
        sm.set_training_values(xt, yt)
        sm.train()

        # DOE for validation
        x = np.linspace(0, 4, 5)
        x2 = np.linspace(-5, 5, 21)
        x1 = []
        for element in itertools.product(x, x2):
            x1.append(np.array(element))
        x_pred = np.array(x1)

        y = sm.predict_values(x_pred)
        yvar = sm.predict_variances(x_pred)

        # prediction are correct on known points
        self.assertTrue(np.abs(np.sum(np.array([y[20], y[50], y[95]]) - yt)) < 1e-6)
        self.assertTrue(np.abs(np.sum(np.array([yvar[20], yvar[50], yvar[95]]))) < 1e-6)

        self.assertEqual(np.shape(y), (105, 1))

    def test_mixed_homo_hyp_2D(self):
        xt = np.array([[0, 5], [2, -1], [4, 0.5]])
        yt = np.array([[0.0], [1.0], [1.5]])
        design_space = DesignSpace(
            [
                CategoricalVariable(["0.0", "1.0", " 2.0", "3.0", "4.0"]),
                FloatVariable(-5, 5),
            ]
        )

        # Surrogate
        sm = MixedIntegerKrigingModel(
            surrogate=KRG(
                design_space=design_space,
                theta0=[1e-2],
                categorical_kernel=MixIntKernelType.HOMO_HSPHERE,
                hyper_opt="Cobyla",
                corr="abs_exp",
            ),
        )
        sm.set_training_values(xt, yt)
        sm.train()

        # DOE for validation
        x = np.linspace(0, 4, 5)
        x2 = np.linspace(-5, 5, 21)
        x1 = []
        for element in itertools.product(x, x2):
            x1.append(np.array(element))
        x_pred = np.array(x1)

        y = sm.predict_values(x_pred)
        yvar = sm.predict_variances(x_pred)

        # prediction are correct on known points
        self.assertTrue(np.abs(np.sum(np.array([y[20], y[50], y[95]]) - yt)) < 1e-6)
        self.assertTrue(np.abs(np.sum(np.array([yvar[20], yvar[50], yvar[95]]))) < 1e-6)

        self.assertEqual(np.shape(y), (105, 1))

    def test_mixed_homo_gaussian_3D_PLS(self):
        xt = np.array([[0.5, 0, 5], [2, 3, 4], [5, 2, -1], [-2, 4, 0.5]])
        yt = np.array([[0.0], [3], [1.0], [1.5]])
        design_space = DesignSpace(
            [
                FloatVariable(-5, 5),
                CategoricalVariable(["0.0", "1.0", " 2.0", "3.0", "4.0"]),
                FloatVariable(-5, 5),
            ]
        )

        # Surrogate
        sm = KPLS(
            design_space=design_space,
            theta0=[1e-2],
            n_comp=1,
            categorical_kernel=MixIntKernelType.EXP_HOMO_HSPHERE,
            cat_kernel_comps=[3],
            corr="squar_exp",
        )
        sm.set_training_values(xt, yt)
        sm.train()

        # DOE for validation
        x = np.linspace(0, 4, 5)
        x2 = np.linspace(-5, 5, 21)
        x1 = []
        for element in itertools.product(x2, x, x2):
            x1.append(np.array(element))
        x_pred = np.array(x1)

        i = 0
        i += 1
        _y = sm.predict_values(x_pred)
        _yvar = sm.predict_variances(x_pred)

        self.assertTrue((np.abs(np.sum(np.array(sm.predict_values(xt) - yt)))) < 1e-6)
        self.assertTrue((np.abs(np.sum(np.array(sm.predict_variances(xt) - 0)))) < 1e-6)

    def test_mixed_homo_gaussian_3D_PLS_cate(self):
        xt = np.array([[0.5, 0, 5], [2, 3, 4], [5, 2, -1], [-2, 4, 0.5]])
        yt = np.array([[0.0], [3], [1.0], [1.5]])
        design_space = DesignSpace(
            [
                FloatVariable(-5, 5),
                CategoricalVariable(["0.0", "1.0", " 2.0", "3.0", "4.0"]),
                FloatVariable(-5, 5),
            ]
        )

        # Surrogate
        sm = KPLS(
            design_space=design_space,
            theta0=[1e-2],
            n_comp=2,
            hyper_opt="Cobyla",
            corr="abs_exp",
            cat_kernel_comps=[3],
            categorical_kernel=MixIntKernelType.EXP_HOMO_HSPHERE,
        )

        sm.set_training_values(xt, yt)
        sm.train()

        # DOE for validation
        x = np.linspace(0, 4, 5)
        x2 = np.linspace(-5, 5, 21)
        x1 = []
        for element in itertools.product(x2, x, x2):
            x1.append(np.array(element))
        x_pred = np.array(x1)

        i = 0
        i += 1
        _y = sm.predict_values(x_pred)
        _yvar = sm.predict_variances(x_pred)

        self.assertTrue((np.abs(np.sum(np.array(sm.predict_values(xt) - yt)))) < 1e-6)
        self.assertTrue((np.abs(np.sum(np.array(sm.predict_variances(xt) - 0)))) < 1e-6)

    def test_mixed_homo_hyp_3D_PLS_cate(self):
        xt = np.array([[0.5, 0, 5], [2, 3, 4], [5, 2, -1], [-2, 4, 0.5]])
        yt = np.array([[0.0], [3], [1.0], [1.5]])
        design_space = DesignSpace(
            [
                FloatVariable(-5, 5),
                CategoricalVariable(["0.0", "1.0", " 2.0", "3.0", "4.0"]),
                FloatVariable(-5, 5),
            ]
        )

        # Surrogate
        sm = MixedIntegerKrigingModel(
            surrogate=KPLS(
                design_space=design_space,
                theta0=[1e-2],
                n_comp=1,
                categorical_kernel=MixIntKernelType.HOMO_HSPHERE,
                cat_kernel_comps=[3],
                hyper_opt="Cobyla",
                corr="squar_exp",
            ),
        )
        sm.set_training_values(xt, yt)
        sm.train()

        # DOE for validation
        x = np.linspace(0, 4, 5)
        x2 = np.linspace(-5, 5, 21)
        x1 = []
        for element in itertools.product(x2, x, x2):
            x1.append(np.array(element))
        x_pred = np.array(x1)

        i = 0
        i += 1
        _y = sm.predict_values(x_pred)
        _yvar = sm.predict_variances(x_pred)

        self.assertTrue((np.abs(np.sum(np.array(sm.predict_values(xt) - yt)))) < 1e-6)
        self.assertTrue((np.abs(np.sum(np.array(sm.predict_variances(xt) - 0)))) < 1e-6)

    def test_compound_hetero_noise_auto(self):
        xt1 = np.array([[0, 0.0], [0, 2.0], [0, 4.0]])
        xt2 = np.array([[1, 0.0], [1, 2.0], [1, 3.0]])
        xt3 = np.array([[2, 1.0], [2, 2.0], [2, 4.0]])

        xt = np.concatenate((xt1, xt2, xt3), axis=0)
        xt[:, 1] = xt[:, 1].astype(np.float64)
        yt1 = np.array([0.0, 9.0, 16.0])
        yt2 = np.array([0.0, -4, -13.0])
        yt3 = np.array([-10, 3, 11.0])
        yt = np.concatenate((yt1, yt2, yt3), axis=0)

        design_space = DesignSpace(
            [
                CategoricalVariable(["Blue", "Red", "Green"]),
                # OrdinalVariable([0,1,2]),
                FloatVariable(0, 4),
            ]
        )

        # Surrogate
        sm = MixedIntegerKrigingModel(
            surrogate=KRG(
                design_space=design_space,
                categorical_kernel=MixIntKernelType.COMPOUND_SYMMETRY,
                theta0=[1e-1],
                corr="squar_exp",
                n_start=20,
                eval_noise=True,
                #     noise0 = [1.0,0.1,10,0.01,1.0,0.1,10,0.01,10],
                use_het_noise=True,
                hyper_opt="Cobyla",
            ),
        )
        sm.set_training_values(xt, yt)
        sm.train()

        # DOE for validation
        n = 100
        x_cat1 = []
        x_cat2 = []
        x_cat3 = []

        for i in range(n):
            x_cat1.append(0)
            x_cat2.append(1)
            x_cat3.append(2)

        x_cont = np.linspace(0.0, 4.0, n)
        x1 = np.concatenate(
            (np.asarray(x_cat1).reshape(-1, 1), x_cont.reshape(-1, 1)), axis=1
        )
        x2 = np.concatenate(
            (np.asarray(x_cat2).reshape(-1, 1), x_cont.reshape(-1, 1)), axis=1
        )
        x3 = np.concatenate(
            (np.asarray(x_cat3).reshape(-1, 1), x_cont.reshape(-1, 1)), axis=1
        )

        _y1 = sm.predict_values(x1)
        _y2 = sm.predict_values(x2)
        _y3 = sm.predict_values(x3)

        # estimated variance
        _s2_1 = sm.predict_variances(x1)
        _s2_2 = sm.predict_variances(x2)
        _s2_3 = sm.predict_variances(x3)

        self.assertEqual(
            np.array_equal(
                np.array([0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0]),
                sm._surrogate.optimal_noise,
            ),
            True,
        )

    def test_mixed_homo_gaussian_3D_ord_cate(self):
        xt = np.array([[0, 5, 0], [2, 4, 3], [4, -1, 2], [2, 0.5, 1]])
        yt = np.array([[0.0], [3], [1.0], [1.5]])
        design_space = DesignSpace(
            [
                CategoricalVariable(["0.0", "1.0", " 2.0", "3.0", "4.0"]),
                FloatVariable(-5, 5),
                CategoricalVariable(["0.0", "1.0", " 2.0", "3.0"]),
            ]
        )

        # Surrogate
        sm = MixedIntegerKrigingModel(
            surrogate=KPLS(
                design_space=design_space,
                theta0=[1e-2],
                n_comp=1,
                categorical_kernel=MixIntKernelType.EXP_HOMO_HSPHERE,
                cat_kernel_comps=[3, 2],
                hyper_opt="Cobyla",
                corr="squar_exp",
            ),
        )
        sm.set_training_values(xt, yt)
        sm.train()

        # DOE for validation
        x = np.linspace(0, 4, 5)
        x2 = np.linspace(-5, 5, 21)
        x3 = np.linspace(0, 3, 4)
        x1 = []
        for element in itertools.product(x, x2, x3):
            x1.append(np.array(element))
        x_pred = np.array(x1)

        _y = sm.predict_values(x_pred)
        _yvar = sm.predict_variances(x_pred)

        # prediction are correct on known points
        self.assertTrue((np.abs(np.sum(np.array(sm.predict_values(xt) - yt)) < 1e-6)))
        self.assertTrue((np.abs(np.sum(np.array(sm.predict_variances(xt) - 0)) < 1e-6)))

    def test_mixed_gower_3D(self):
        design_space = DesignSpace(
            [
                FloatVariable(-10, 10),
                IntegerVariable(-10, 10),
                IntegerVariable(-10, 10),
            ]
        )
        mixint = MixedIntegerContext(design_space)

        sm = mixint.build_kriging_model(
            KRG(
                categorical_kernel=MixIntKernelType.GOWER,
                print_prediction=False,
                hyper_opt="Cobyla",
            )
        )
        sampling = mixint.build_sampling_method()

        fun = Sphere(ndim=3)
        xt = sampling(10)
        yt = fun(xt)
        sm.set_training_values(xt, yt)
        sm.train()
        eq_check = True
        for i in range(xt.shape[0]):
            if abs(float(xt[i, :][1]) - int(float(xt[i, :][1]))) > 10e-8:
                eq_check = False
        self.assertTrue(eq_check)

    def run_mixed_gower_example(self):
        import matplotlib.pyplot as plt
        import numpy as np

        from smt.applications.mixed_integer import (
            MixedIntegerKrigingModel,
        )
        from smt.design_space import (
            CategoricalVariable,
            DesignSpace,
            FloatVariable,
        )
        from smt.surrogate_models import KRG, MixIntKernelType

        xt1 = np.array([[0, 0.0], [0, 2.0], [0, 4.0]])
        xt2 = np.array([[1, 0.0], [1, 2.0], [1, 3.0]])
        xt3 = np.array([[2, 1.0], [2, 2.0], [2, 4.0]])

        xt = np.concatenate((xt1, xt2, xt3), axis=0)
        xt[:, 1] = xt[:, 1].astype(np.float64)
        yt1 = np.array([0.0, 9.0, 16.0])
        yt2 = np.array([0.0, -4, -13.0])
        yt3 = np.array([-10, 3, 11.0])
        yt = np.concatenate((yt1, yt2, yt3), axis=0)

        design_space = DesignSpace(
            [
                CategoricalVariable(["Blue", "Red", "Green"]),
                FloatVariable(0, 4),
            ]
        )

        # Surrogate
        sm = MixedIntegerKrigingModel(
            surrogate=KRG(
                design_space=design_space,
                categorical_kernel=MixIntKernelType.GOWER,
                theta0=[1e-1],
                hyper_opt="Cobyla",
                corr="squar_exp",
                n_start=20,
            ),
        )
        sm.set_training_values(xt, yt)
        sm.train()

        # DOE for validation
        n = 100
        x_cat1 = []
        x_cat2 = []
        x_cat3 = []

        for i in range(n):
            x_cat1.append(0)
            x_cat2.append(1)
            x_cat3.append(2)

        x_cont = np.linspace(0.0, 4.0, n)
        x1 = np.concatenate(
            (np.asarray(x_cat1).reshape(-1, 1), x_cont.reshape(-1, 1)), axis=1
        )
        x2 = np.concatenate(
            (np.asarray(x_cat2).reshape(-1, 1), x_cont.reshape(-1, 1)), axis=1
        )
        x3 = np.concatenate(
            (np.asarray(x_cat3).reshape(-1, 1), x_cont.reshape(-1, 1)), axis=1
        )

        y1 = sm.predict_values(x1)
        y2 = sm.predict_values(x2)
        y3 = sm.predict_values(x3)

        # estimated variance
        s2_1 = sm.predict_variances(x1)
        s2_2 = sm.predict_variances(x2)
        s2_3 = sm.predict_variances(x3)

        fig, axs = plt.subplots(3, figsize=(8, 6))

        axs[0].plot(xt1[:, 1].astype(np.float64), yt1, "o", linestyle="None")
        axs[0].plot(x_cont, y1, color="Blue")
        axs[0].fill_between(
            np.ravel(x_cont),
            np.ravel(y1 - 3 * np.sqrt(s2_1)),
            np.ravel(y1 + 3 * np.sqrt(s2_1)),
            color="lightgrey",
        )
        axs[0].set_xlabel("x")
        axs[0].set_ylabel("y")
        axs[0].legend(
            ["Training data", "Prediction", "Confidence Interval 99%"],
            loc="upper left",
            bbox_to_anchor=[0, 1],
        )
        axs[1].plot(
            xt2[:, 1].astype(np.float64), yt2, marker="o", color="r", linestyle="None"
        )
        axs[1].plot(x_cont, y2, color="Red")
        axs[1].fill_between(
            np.ravel(x_cont),
            np.ravel(y2 - 3 * np.sqrt(s2_2)),
            np.ravel(y2 + 3 * np.sqrt(s2_2)),
            color="lightgrey",
        )
        axs[1].set_xlabel("x")
        axs[1].set_ylabel("y")
        axs[1].legend(
            ["Training data", "Prediction", "Confidence Interval 99%"],
            loc="upper left",
            bbox_to_anchor=[0, 1],
        )
        axs[2].plot(
            xt3[:, 1].astype(np.float64), yt3, marker="o", color="r", linestyle="None"
        )
        axs[2].plot(x_cont, y3, color="Green")
        axs[2].fill_between(
            np.ravel(x_cont),
            np.ravel(y3 - 3 * np.sqrt(s2_3)),
            np.ravel(y3 + 3 * np.sqrt(s2_3)),
            color="lightgrey",
        )
        axs[2].set_xlabel("x")
        axs[2].set_ylabel("y")
        axs[2].legend(
            ["Training data", "Prediction", "Confidence Interval 99%"],
            loc="upper left",
            bbox_to_anchor=[0, 1],
        )
        plt.tight_layout()
        plt.show()

    # FIXME: Used in SMT documentation but belongs to smt_design_space_ext domain
    def run_mixed_cs_example(self):
        import matplotlib.pyplot as plt
        import numpy as np

        from smt.applications.mixed_integer import (
            MixedIntegerKrigingModel,
        )
        from smt.design_space import (
            CategoricalVariable,
            DesignSpace,
            FloatVariable,
        )
        from smt.surrogate_models import KRG, MixIntKernelType

        xt1 = np.array([[0, 0.0], [0, 2.0], [0, 4.0]])
        xt2 = np.array([[1, 0.0], [1, 2.0], [1, 3.0]])
        xt3 = np.array([[2, 1.0], [2, 2.0], [2, 4.0]])

        xt = np.concatenate((xt1, xt2, xt3), axis=0)
        xt[:, 1] = xt[:, 1].astype(np.float64)
        yt1 = np.array([0.0, 9.0, 16.0])
        yt2 = np.array([0.0, -4, -13.0])
        yt3 = np.array([-10, 3, 11.0])
        yt = np.concatenate((yt1, yt2, yt3), axis=0)

        design_space = DesignSpace(
            [
                CategoricalVariable(["Blue", "Red", "Green"]),
                FloatVariable(0, 4),
            ]
        )

        # Surrogate
        sm = MixedIntegerKrigingModel(
            surrogate=KRG(
                design_space=design_space,
                categorical_kernel=MixIntKernelType.COMPOUND_SYMMETRY,
                theta0=[1e-1],
                hyper_opt="Cobyla",
                corr="squar_exp",
                n_start=20,
            ),
        )
        sm.set_training_values(xt, yt)
        sm.train()

        # DOE for validation
        n = 100
        x_cat1 = []
        x_cat2 = []
        x_cat3 = []

        for i in range(n):
            x_cat1.append(0)
            x_cat2.append(1)
            x_cat3.append(2)

        x_cont = np.linspace(0.0, 4.0, n)
        x1 = np.concatenate(
            (np.asarray(x_cat1).reshape(-1, 1), x_cont.reshape(-1, 1)), axis=1
        )
        x2 = np.concatenate(
            (np.asarray(x_cat2).reshape(-1, 1), x_cont.reshape(-1, 1)), axis=1
        )
        x3 = np.concatenate(
            (np.asarray(x_cat3).reshape(-1, 1), x_cont.reshape(-1, 1)), axis=1
        )

        y1 = sm.predict_values(x1)
        y2 = sm.predict_values(x2)
        y3 = sm.predict_values(x3)

        # estimated variance
        s2_1 = sm.predict_variances(x1)
        s2_2 = sm.predict_variances(x2)
        s2_3 = sm.predict_variances(x3)

        fig, axs = plt.subplots(3, figsize=(8, 6))

        axs[0].plot(xt1[:, 1].astype(np.float64), yt1, "o", linestyle="None")
        axs[0].plot(x_cont, y1, color="Blue")
        axs[0].fill_between(
            np.ravel(x_cont),
            np.ravel(y1 - 3 * np.sqrt(s2_1)),
            np.ravel(y1 + 3 * np.sqrt(s2_1)),
            color="lightgrey",
        )
        axs[0].set_xlabel("x")
        axs[0].set_ylabel("y")
        axs[0].legend(
            ["Training data", "Prediction", "Confidence Interval 99%"],
            loc="upper left",
            bbox_to_anchor=[0, 1],
        )
        axs[1].plot(
            xt2[:, 1].astype(np.float64), yt2, marker="o", color="r", linestyle="None"
        )
        axs[1].plot(x_cont, y2, color="Red")
        axs[1].fill_between(
            np.ravel(x_cont),
            np.ravel(y2 - 3 * np.sqrt(s2_2)),
            np.ravel(y2 + 3 * np.sqrt(s2_2)),
            color="lightgrey",
        )
        axs[1].set_xlabel("x")
        axs[1].set_ylabel("y")
        axs[1].legend(
            ["Training data", "Prediction", "Confidence Interval 99%"],
            loc="upper left",
            bbox_to_anchor=[0, 1],
        )
        axs[2].plot(
            xt3[:, 1].astype(np.float64), yt3, marker="o", color="r", linestyle="None"
        )
        axs[2].plot(x_cont, y3, color="Green")
        axs[2].fill_between(
            np.ravel(x_cont),
            np.ravel(y3 - 3 * np.sqrt(s2_3)),
            np.ravel(y3 + 3 * np.sqrt(s2_3)),
            color="lightgrey",
        )
        axs[2].set_xlabel("x")
        axs[2].set_ylabel("y")
        axs[2].legend(
            ["Training data", "Prediction", "Confidence Interval 99%"],
            loc="upper left",
            bbox_to_anchor=[0, 1],
        )
        plt.tight_layout()
        plt.show()

    # FIXME: Used in SMT documentation but belongs to smt_design_space_ext domain
    def run_mixed_homo_gaussian_example(self):
        import matplotlib.pyplot as plt
        import numpy as np

        from smt.applications.mixed_integer import MixedIntegerKrigingModel
        from smt.design_space import (
            CategoricalVariable,
            DesignSpace,
            FloatVariable,
        )
        from smt.surrogate_models import KRG, MixIntKernelType

        xt1 = np.array([[0, 0.0], [0, 2.0], [0, 4.0]])
        xt2 = np.array([[1, 0.0], [1, 2.0], [1, 3.0]])
        xt3 = np.array([[2, 1.0], [2, 2.0], [2, 4.0]])

        xt = np.concatenate((xt1, xt2, xt3), axis=0)
        xt[:, 1] = xt[:, 1].astype(np.float64)
        yt1 = np.array([0.0, 9.0, 16.0])
        yt2 = np.array([0.0, -4, -13.0])
        yt3 = np.array([-10, 3, 11.0])
        yt = np.concatenate((yt1, yt2, yt3), axis=0)

        design_space = DesignSpace(
            [
                CategoricalVariable(["Blue", "Red", "Green"]),
                FloatVariable(0, 4),
            ]
        )

        # Surrogate
        sm = MixedIntegerKrigingModel(
            surrogate=KRG(
                design_space=design_space,
                theta0=[1e-1],
                hyper_opt="Cobyla",
                corr="squar_exp",
                n_start=20,
                categorical_kernel=MixIntKernelType.EXP_HOMO_HSPHERE,
            ),
        )
        sm.set_training_values(xt, yt)
        sm.train()

        # DOE for validation
        n = 100
        x_cat1 = []
        x_cat2 = []
        x_cat3 = []

        for i in range(n):
            x_cat1.append(0)
            x_cat2.append(1)
            x_cat3.append(2)

        x_cont = np.linspace(0.0, 4.0, n)
        x1 = np.concatenate(
            (np.asarray(x_cat1).reshape(-1, 1), x_cont.reshape(-1, 1)), axis=1
        )
        x2 = np.concatenate(
            (np.asarray(x_cat2).reshape(-1, 1), x_cont.reshape(-1, 1)), axis=1
        )
        x3 = np.concatenate(
            (np.asarray(x_cat3).reshape(-1, 1), x_cont.reshape(-1, 1)), axis=1
        )

        y1 = sm.predict_values(x1)
        y2 = sm.predict_values(x2)
        y3 = sm.predict_values(x3)

        # estimated variance
        s2_1 = sm.predict_variances(x1)
        s2_2 = sm.predict_variances(x2)
        s2_3 = sm.predict_variances(x3)

        fig, axs = plt.subplots(3, figsize=(8, 6))

        axs[0].plot(xt1[:, 1].astype(np.float64), yt1, "o", linestyle="None")
        axs[0].plot(x_cont, y1, color="Blue")
        axs[0].fill_between(
            np.ravel(x_cont),
            np.ravel(y1 - 3 * np.sqrt(s2_1)),
            np.ravel(y1 + 3 * np.sqrt(s2_1)),
            color="lightgrey",
        )
        axs[0].set_xlabel("x")
        axs[0].set_ylabel("y")
        axs[0].legend(
            ["Training data", "Prediction", "Confidence Interval 99%"],
            loc="upper left",
            bbox_to_anchor=[0, 1],
        )
        axs[1].plot(
            xt2[:, 1].astype(np.float64), yt2, marker="o", color="r", linestyle="None"
        )
        axs[1].plot(x_cont, y2, color="Red")
        axs[1].fill_between(
            np.ravel(x_cont),
            np.ravel(y2 - 3 * np.sqrt(s2_2)),
            np.ravel(y2 + 3 * np.sqrt(s2_2)),
            color="lightgrey",
        )
        axs[1].set_xlabel("x")
        axs[1].set_ylabel("y")
        axs[1].legend(
            ["Training data", "Prediction", "Confidence Interval 99%"],
            loc="upper left",
            bbox_to_anchor=[0, 1],
        )
        axs[2].plot(
            xt3[:, 1].astype(np.float64), yt3, marker="o", color="r", linestyle="None"
        )
        axs[2].plot(x_cont, y3, color="Green")
        axs[2].fill_between(
            np.ravel(x_cont),
            np.ravel(y3 - 3 * np.sqrt(s2_3)),
            np.ravel(y3 + 3 * np.sqrt(s2_3)),
            color="lightgrey",
        )
        axs[2].set_xlabel("x")
        axs[2].set_ylabel("y")
        axs[2].legend(
            ["Training data", "Prediction", "Confidence Interval 99%"],
            loc="upper left",
            bbox_to_anchor=[0, 1],
        )
        plt.tight_layout()
        plt.show()

    def run_mixed_homo_hyp_example(self):
        import matplotlib.pyplot as plt
        import numpy as np

        from smt.applications.mixed_integer import MixedIntegerKrigingModel
        from smt.design_space import (
            CategoricalVariable,
            DesignSpace,
            FloatVariable,
        )
        from smt.surrogate_models import KRG, MixIntKernelType

        xt1 = np.array([[0, 0.0], [0, 2.0], [0, 4.0]])
        xt2 = np.array([[1, 0.0], [1, 2.0], [1, 3.0]])
        xt3 = np.array([[2, 1.0], [2, 2.0], [2, 4.0]])

        xt = np.concatenate((xt1, xt2, xt3), axis=0)
        xt[:, 1] = xt[:, 1].astype(np.float64)
        yt1 = np.array([0.0, 9.0, 16.0])
        yt2 = np.array([0.0, -4, -13.0])
        yt3 = np.array([-10, 3, 11.0])
        yt = np.concatenate((yt1, yt2, yt3), axis=0)

        design_space = DesignSpace(
            [
                CategoricalVariable(["Blue", "Red", "Green"]),
                FloatVariable(0, 4),
            ]
        )

        # Surrogate
        sm = MixedIntegerKrigingModel(
            surrogate=KRG(
                design_space=design_space,
                categorical_kernel=MixIntKernelType.HOMO_HSPHERE,
                theta0=[1e-1],
                hyper_opt="Cobyla",
                corr="squar_exp",
                n_start=20,
            ),
        )
        sm.set_training_values(xt, yt)
        sm.train()

        # DOE for validation
        n = 100
        x_cat1 = []
        x_cat2 = []
        x_cat3 = []

        for i in range(n):
            x_cat1.append(0)
            x_cat2.append(1)
            x_cat3.append(2)

        x_cont = np.linspace(0.0, 4.0, n)
        x1 = np.concatenate(
            (np.asarray(x_cat1).reshape(-1, 1), x_cont.reshape(-1, 1)), axis=1
        )
        x2 = np.concatenate(
            (np.asarray(x_cat2).reshape(-1, 1), x_cont.reshape(-1, 1)), axis=1
        )
        x3 = np.concatenate(
            (np.asarray(x_cat3).reshape(-1, 1), x_cont.reshape(-1, 1)), axis=1
        )

        y1 = sm.predict_values(x1)
        y2 = sm.predict_values(x2)
        y3 = sm.predict_values(x3)

        # estimated variance
        s2_1 = sm.predict_variances(x1)
        s2_2 = sm.predict_variances(x2)
        s2_3 = sm.predict_variances(x3)

        fig, axs = plt.subplots(3, figsize=(8, 6))

        axs[0].plot(xt1[:, 1].astype(np.float64), yt1, "o", linestyle="None")
        axs[0].plot(x_cont, y1, color="Blue")
        axs[0].fill_between(
            np.ravel(x_cont),
            np.ravel(y1 - 3 * np.sqrt(s2_1)),
            np.ravel(y1 + 3 * np.sqrt(s2_1)),
            color="lightgrey",
        )
        axs[0].set_xlabel("x")
        axs[0].set_ylabel("y")
        axs[0].legend(
            ["Training data", "Prediction", "Confidence Interval 99%"],
            loc="upper left",
            bbox_to_anchor=[0, 1],
        )
        axs[1].plot(
            xt2[:, 1].astype(np.float64), yt2, marker="o", color="r", linestyle="None"
        )
        axs[1].plot(x_cont, y2, color="Red")
        axs[1].fill_between(
            np.ravel(x_cont),
            np.ravel(y2 - 3 * np.sqrt(s2_2)),
            np.ravel(y2 + 3 * np.sqrt(s2_2)),
            color="lightgrey",
        )
        axs[1].set_xlabel("x")
        axs[1].set_ylabel("y")
        axs[1].legend(
            ["Training data", "Prediction", "Confidence Interval 99%"],
            loc="upper left",
            bbox_to_anchor=[0, 1],
        )
        axs[2].plot(
            xt3[:, 1].astype(np.float64), yt3, marker="o", color="r", linestyle="None"
        )
        axs[2].plot(x_cont, y3, color="Green")
        axs[2].fill_between(
            np.ravel(x_cont),
            np.ravel(y3 - 3 * np.sqrt(s2_3)),
            np.ravel(y3 + 3 * np.sqrt(s2_3)),
            color="lightgrey",
        )
        axs[2].set_xlabel("x")
        axs[2].set_ylabel("y")
        axs[2].legend(
            ["Training data", "Prediction", "Confidence Interval 99%"],
            loc="upper left",
            bbox_to_anchor=[0, 1],
        )
        plt.tight_layout()
        plt.show()


if __name__ == "__main__":
    TestMixedInteger().run_mixed_integer_context_example()
    unittest.main()
